# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import os
import platform
import shutil
import time
import warnings

import torch

from .base_runner import BaseRunner
from .builder import RUNNERS
from .checkpoint import save_checkpoint
from .utils import get_host_info


@RUNNERS.register_module()
class EpochBasedRunner(BaseRunner):
    """Epoch-based Runner.

    This runner train models epoch by epoch.
    """

    def run_iter(self, data_batch, train_mode, **kwargs):
        if train_mode:
            outputs = self.model.train_step(data_batch, self.optimizer, **kwargs)
        else:
            outputs = self.model.val_step(data_batch, self.optimizer, **kwargs)

        if not isinstance(outputs, dict):
            raise TypeError('"model.train_step()"'
                            'and "model.val_step()" must return a dict')

        if 'log_vars' in outputs:
            self.log_buffer.update(outputs['log_vars'], outputs['num_samples'])
        self.outputs = outputs

    def train(self, data_loader, **kwargs):
        self.model.train()
        self.mode = 'train'
        self.data_loader = data_loader
        self._max_iters = self._max_epochs * len(self.data_loader)
        self.call_hook('before_train_epoch')

        time.sleep(2)  # Prevent possible deadlock during epoch transition
        for i, data_batch in enumerate(self.data_loader):
            self._inner_iter = i
            self.call_hook('before_train_iter')
            kwargs.update({'epoch': self._epoch, 'iter': self._iter})
            self.run_iter(data_batch, train_mode=True, **kwargs)
            self.call_hook('after_train_iter')
            self._iter += 1

        self.call_hook('after_train_epoch')
        self._epoch += 1
        # 更新data_loader中dataset记录的epoch计数
        self.data_loader.dataset.epoch = self._epoch    # 仅在iterator dataset数据集有效
        

    @torch.no_grad()
    def val(self, data_loader, **kwargs):
        self.model.eval()
        self.mode = 'val'
        self.data_loader = data_loader
        self.call_hook('before_val_epoch')
        time.sleep(2)  # Prevent possible deadlock during epoch transition
        for i, data_batch in enumerate(self.data_loader):
            self._inner_iter = i
            self.call_hook('before_val_iter')
            self.run_iter(data_batch, train_mode=False, **kwargs)
            self.call_hook('after_val_iter')

        self.call_hook('after_val_epoch')

    def run(self, data_loaders, workflow, max_epochs, **kwargs):
        """Start running.

        Args:
            data_loaders (list[:obj:`DataLoader`]): Dataloaders for training
                and validation.
            workflow (list[tuple]): A list of (phase, epochs) to specify the
                running order and epochs. E.g, [('train', 2), ('val', 1)] means
                running 2 epochs for training and 1 epoch for validation,
                iteratively.
        """
        assert isinstance(data_loaders, list)
        assert len(data_loaders) == len(workflow)
        self._max_epochs = max_epochs

        work_dir = self.work_dir if self.work_dir is not None else 'NONE'
        self.logger.info('Start running, host: %s, work_dir: %s',
                         get_host_info(), work_dir)
        self.logger.info('Hooks will be executed in the following order:\n%s',
                         self.get_hook_info())
        self.logger.info('workflow: %s, max: %d epochs', workflow, self._max_epochs)
        self.call_hook('before_run')

        while self.epoch < self._max_epochs:
            for i, flow in enumerate(workflow):
                mode, epochs = flow
                if isinstance(mode, str):  # self.train()
                    if not hasattr(self, mode):
                        raise ValueError(
                            f'runner has no method named "{mode}" to run an '
                            'epoch')
                    epoch_runner = getattr(self, mode)
                else:
                    raise TypeError(
                        'mode in workflow must be a str, but got {}'.format(
                            type(mode)))

                for _ in range(epochs):
                    if mode == 'train' and self.epoch >= self._max_epochs:
                        break
                    epoch_runner(data_loaders[i], **kwargs)

        time.sleep(1)  # wait for some hooks like loggers to finish
        self.call_hook('after_run')

    def save_checkpoint(self,
                        out_dir,
                        filename_tmpl='epoch_{}.pth',
                        save_optimizer=True,
                        meta=None,
                        create_symlink=True):
        """Save the checkpoint.

        Args:
            out_dir (str): The directory that checkpoints are saved.
            filename_tmpl (str, optional): The checkpoint filename template,
                which contains a placeholder for the epoch number.
                Defaults to 'epoch_{}.pth'.
            save_optimizer (bool, optional): Whether to save the optimizer to
                the checkpoint. Defaults to True.
            meta (dict, optional): The meta information to be saved in the
                checkpoint. Defaults to None.
            create_symlink (bool, optional): Whether to create a symlink
                "latest.pth" to point to the latest checkpoint.
                Defaults to True.
        """
        if meta is None:
            meta = {}
        elif not isinstance(meta, dict):
            raise TypeError(
                f'meta should be a dict or None, but got {type(meta)}')
        if self.meta is not None:
            meta.update(self.meta)
            # Note: meta.update(self.meta) should be done before
            # meta.update(epoch=self.epoch + 1, iter=self.iter) otherwise
            # there will be problems with resumed checkpoints.
            # More details in https://github.com/open-mmlab/mmcv/pull/1108
        meta.update(epoch=self.epoch + 1, iter=self.iter)

        filename = filename_tmpl.format(self.epoch + 1)
        filepath = osp.join(out_dir, filename)
        optimizer = self.optimizer if save_optimizer else None
        save_checkpoint(self.model, filepath, optimizer=optimizer, meta=meta)
        # in some environments, `os.symlink` is not supported, you may need to
        # set `create_symlink` to False
        if create_symlink:
            dst_file = osp.join(out_dir, 'latest.pth')
            if platform.system() != 'Windows':
                if os.path.exists(dst_file):
                    os.remove(dst_file)
                os.symlink(filename, dst_file)
            else:
                shutil.copy(filepath, dst_file)



import copy
import warnings
from abc import ABCMeta
from collections import defaultdict
from logging import FileHandler
import torch.distributed as dist
import torch.nn as nn
from collections import OrderedDict
import torch
from antgo.framework.helper.utils.logging import logger_initialized, print_log


class BaseModule(nn.Module, metaclass=ABCMeta):
    """Base module for all modules in openmmlab.

    ``BaseModule`` is a wrapper of ``torch.nn.Module`` with additional
    functionality of parameter initialization. Compared with
    ``torch.nn.Module``, ``BaseModule`` mainly adds three attributes.

    - ``init_cfg``: the config to control the initialization.
    - ``init_weights``: The function of parameter initialization and recording
      initialization information.
    - ``_params_init_info``: Used to track the parameter initialization
      information. This attribute only exists during executing the
      ``init_weights``.

    Args:
        init_cfg (dict, optional): Initialization config dict.
    """
    def __init__(self, init_cfg=None):
        """Initialize BaseModule, inherited from `torch.nn.Module`"""

        # NOTE init_cfg can be defined in different levels, but init_cfg
        # in low levels has a higher priority.

        super(BaseModule, self).__init__()
        # define default value of init_cfg instead of hard code
        # in init_weights() function
        self._is_init = False

        self.init_cfg = copy.deepcopy(init_cfg)

    @property
    def is_init(self):
        return self._is_init

    def init_weights(self):
        """Initialize the weights."""

        is_top_level_module = False
        # check if it is top-level module
        if not hasattr(self, '_params_init_info'):
            # The `_params_init_info` is used to record the initialization
            # information of the parameters
            # the key should be the obj:`nn.Parameter` of model and the value
            # should be a dict containing
            # - init_info (str): The string that describes the initialization.
            # - tmp_mean_value (FloatTensor): The mean of the parameter,
            #       which indicates whether the parameter has been modified.
            # this attribute would be deleted after all parameters
            # is initialized.
            self._params_init_info = defaultdict(dict)
            is_top_level_module = True

            # Initialize the `_params_init_info`,
            # When detecting the `tmp_mean_value` of
            # the corresponding parameter is changed, update related
            # initialization information
            for name, param in self.named_parameters():
                self._params_init_info[param][
                    'init_info'] = f'The value is the same before and ' \
                                   f'after calling `init_weights` ' \
                                   f'of {self.__class__.__name__} '
                self._params_init_info[param][
                    'tmp_mean_value'] = param.data.mean()

            # pass `params_init_info` to all submodules
            # All submodules share the same `params_init_info`,
            # so it will be updated when parameters are
            # modified at any level of the model.
            for sub_module in self.modules():
                sub_module._params_init_info = self._params_init_info

        # Get the initialized logger, if not exist,
        logger_names = list(logger_initialized.keys())
        logger_name = logger_names[0] if logger_names else 'antgo'

        from antgo.framework.helper.cnn import initialize
        from antgo.framework.helper.cnn.utils.weight_init import update_init_info
        module_name = self.__class__.__name__
        if not self._is_init:
            if self.init_cfg:
                print_log(
                    f'initialize {module_name} with init_cfg {self.init_cfg}',
                    logger=logger_name)
                initialize(self, self.init_cfg)
                if isinstance(self.init_cfg, dict):
                    # prevent the parameters of
                    # the pre-trained model
                    # from being overwritten by
                    # the `init_weights`
                    if self.init_cfg['type'] == 'Pretrained':
                        return

            for m in self.children():
                if hasattr(m, 'init_weights'):
                    m.init_weights()
                    # users may overload the `init_weights`
                    update_init_info(
                        m,
                        init_info=f'Initialized by '
                        f'user-defined `init_weights`'
                        f' in {m.__class__.__name__} ')

            self._is_init = True
        else:
            warnings.warn(f'init_weights of {self.__class__.__name__} has '
                          f'been called more than once.')

        if is_top_level_module:
            for sub_module in self.modules():
                del sub_module._params_init_info

    def _parse_losses(self, losses):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor \
                which may be a weighted sum of all losses, log_vars contains \
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            else:
                raise TypeError(
                    f'{loss_name} is not a tensor or list of tensors')

        # loss = sum(_value for _key, _value in log_vars.items())
        loss = sum(_value for _key, _value in log_vars.items()
                   if 'loss' in _key)

        # If the loss_vars has different length, GPUs will wait infinitely
        if dist.is_available() and dist.is_initialized():
            log_var_length = torch.tensor(len(log_vars), device=loss.device)
            dist.all_reduce(log_var_length)
            message = (f'rank {dist.get_rank()}' +
                       f' len(log_vars): {len(log_vars)}' + ' keys: ' +
                       ','.join(log_vars.keys()))
            assert log_var_length == len(log_vars) * dist.get_world_size(), \
                'loss log variables are different across GPUs!\n' + message

        log_vars['loss'] = loss
        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars

    def train_step(self, data, optimizer, **kwargs):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``, \
                ``num_samples``.

                - ``loss`` is a tensor for back propagation, which can be a
                  weighted sum of multiple losses.
                - ``log_vars`` contains all the variables to be sent to the
                  logger.
                - ``num_samples`` indicates the batch size (when the model is
                  DDP, it means the batch size on each GPU), which is used for
                  averaging the logs.
        """
        num_samples = 1
        losses = None
        if type(data) == list or type(data) == tuple:
            num_samples = len(data[0]['image'])
            losses = self(*data, **kwargs)
        else:
            num_samples = len(data['image'])
            losses = self(**data, **kwargs)

        loss, log_vars = self._parse_losses(losses)
        outputs = dict(
            loss=loss, log_vars=log_vars, num_samples=num_samples)

        return outputs

    def val_step(self, data, optimizer=None, **kwargs):
        """The iteration step during validation.

        This method shares the same signature as :func:`train_step`, but used
        during val epochs. Note that the evaluation after training epochs is
        not implemented with this method, but an evaluation hook.
        """
        num_samples = 1
        losses = None
        if type(data) == list or type(data) == tuple:
            num_samples = len(data[0]['image'])
            losses = self(*data, **kwargs)
        else:
            num_samples = len(data['image'])
            losses = self(**data, **kwargs)

        return_loss = kwargs.get('return_loss', False)

        log_vars = {}
        for loss_name, loss_value in losses.items():
            if return_loss and 'loss' in loss_name:
                if isinstance(loss_value, torch.Tensor):
                    log_vars[loss_name] = loss_value.mean()
                elif isinstance(loss_value, list):
                    log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
                else:
                    raise TypeError(
                        f'{loss_name} is not a tensor or list of tensors')

        loss = 0
        if return_loss:
            loss = sum(_value for _key, _value in log_vars.items()
                    if 'loss' in _key)
            log_vars['loss'] = loss

        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            # TODO, 需要进行多节点损失的reduce?
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        outputs = dict(
            loss=loss, log_vars=log_vars, num_samples=num_samples)

        return outputs

    def forward(self, image, return_loss=True, **kwargs):
        """Calls either :func:`forward_train` or :func:`forward_test` depending
        on whether ``return_loss`` is ``True``.

        Note this setting will change the expected inputs. When
        ``return_loss=True``, img and img_meta are single-nested (i.e. Tensor
        and List[dict]), and when ``resturn_loss=False``, img and img_meta
        should be double nested (i.e.  List[Tensor], List[List[dict]]), with
        the outer list indicating test time augmentations.
        """
        if return_loss:
            return self.forward_train(image, **kwargs)
        else:
            return self.forward_test(image, **kwargs)

    def forward_train(self, image, **kwargs):
        raise NotImplementedError

    def forward_test(self, image, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        s = super().__repr__()
        if self.init_cfg:
            s += f'\ninit_cfg={self.init_cfg}'
        return s


class Sequential(BaseModule, nn.Sequential):
    """Sequential module in openmmlab.

    Args:
        init_cfg (dict, optional): Initialization config dict.
    """

    def __init__(self, *args, init_cfg=None):
        BaseModule.__init__(self, init_cfg)
        nn.Sequential.__init__(self, *args)


class ModuleList(BaseModule, nn.ModuleList):
    """ModuleList in openmmlab.

    Args:
        modules (iterable, optional): an iterable of modules to add.
        init_cfg (dict, optional): Initialization config dict.
    """

    def __init__(self, modules=None, init_cfg=None):
        BaseModule.__init__(self, init_cfg)
        nn.ModuleList.__init__(self, modules)


class ModuleDict(BaseModule, nn.ModuleDict):
    """ModuleDict in openmmlab.

    Args:
        modules (dict, optional): a mapping (dictionary) of (string: module)
            or an iterable of key-value pairs of type (string, module).
        init_cfg (dict, optional): Initialization config dict.
    """

    def __init__(self, modules=None, init_cfg=None):
        BaseModule.__init__(self, init_cfg)
        nn.ModuleDict.__init__(self, modules)



# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

from .acc_dataset import *
from .acc_predictor import *
from .arch_encoder import *



import functools
from ..builder import PIPELINES
import torchvision
import inspect

import torchvision.transforms as transforms
from collections.abc import Sequence
from antgo.framework.helper.utils.registry import build_from_cfg
from antgo.dataflow import imgaug as local_imgaug

def __transform_init__(self, keys, **kwargs):
    self.keys = keys
    self.parent_cls.__init__(self, **kwargs)


def __transform_call__(self, samples):
    for key in self.keys:
        samples[key] = self.parent_cls.__call__(self, samples[key])
    return samples
    

def register_torchvision_transforms():
    torch_transfoms = []
    for module_name in dir(torchvision.transforms):
        if module_name.startswith('__'):
            continue
        
        __transform = getattr(torchvision.transforms, module_name)
        if inspect.isclass(__transform) and module_name not in ['AutoAugmentPolicy', 'InterpolationMode']:
            __transform_proxy = \
                type(
                    module_name, 
                    (__transform,), 
                    {   
                        'name': module_name,
                        '__doc__': f'{module_name}',
                        '__init__':  __transform_init__,
                        '__call__': __transform_call__,
                        'parent_cls': __transform
                    }
                )
            PIPELINES.register_module()(__transform_proxy)
            torch_transfoms.append(module_name)


TORCH_VISION_TRANSFORMS = register_torchvision_transforms()


def register_antgo_pipeline():
    for imgaug_module_name in local_imgaug.__all__:
        PIPELINES.register_module()(getattr(local_imgaug, imgaug_module_name))

ANTGO_TRANSFORMS = register_antgo_pipeline()


# @PIPELINES.register_module()
# class TorchVisionCompose(transforms.Compose):
#     def __init__(self, processes):
#         process_t = []
#         for pn, pp in processes:
#             process_t.append(getattr(transforms, pn)(**pp))

#         super().__init__(process_t)
    
#     def __call__(self, sample):
#         image = sample['image']
#         sample['image'] = super(TorchVisionCompose, self).__call__(image)
#         return sample


# @PIPELINES.register_module()
# class Compose(object):
#     """Compose a data pipeline with a sequence of transforms.

#     Args:
#         transforms (list[dict | callable]):
#             Either config dicts of transforms or transform objects.
#     """

#     def __init__(self, transforms):
#         assert isinstance(transforms, Sequence)
#         self.transforms = []
#         for transform in transforms:
#             if isinstance(transform, dict):
#                 transform = build_from_cfg(transform, PIPELINES)
#                 self.transforms.append(transform)
#             elif callable(transform):
#                 self.transforms.append(transform)
#             else:
#                 raise TypeError('transform must be callable or a dict, but got'
#                                 f' {type(transform)}')

#     def __call__(self, data):
#         for t in self.transforms:
#             data = t(data)
#             if data is None:
#                 return None
#         return data

#     def __repr__(self):
#         format_string = self.__class__.__name__ + '('
#         for t in self.transforms:
#             format_string += f'\n    {t}'
#         format_string += '\n)'
#         return format_string


# Copyright (c) OpenMMLab. All rights reserved.
import warnings
import numpy as np
import torch

from ..builder import BBOX_CODERS
from .base_bbox_coder import BaseBBoxCoder


@BBOX_CODERS.register_module()
class DeltaXYWHBBoxCoder(BaseBBoxCoder):
    """Delta XYWH BBox coder.

    Following the practice in `R-CNN <https://arxiv.org/abs/1311.2524>`_,
    this coder encodes bbox (x1, y1, x2, y2) into delta (dx, dy, dw, dh) and
    decodes delta (dx, dy, dw, dh) back to original bbox (x1, y1, x2, y2).

    Args:
        target_means (Sequence[float]): Denormalizing means of target for
            delta coordinates
        target_stds (Sequence[float]): Denormalizing standard deviation of
            target for delta coordinates
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.
        add_ctr_clamp (bool): Whether to add center clamp, when added, the
            predicted box is clamped is its center is too far away from
            the original anchor's center. Only used by YOLOF. Default False.
        ctr_clamp (int): the maximum pixel shift to clamp. Only used by YOLOF.
            Default 32.
    """

    def __init__(self,
                 target_means=(0., 0., 0., 0.),
                 target_stds=(1., 1., 1., 1.),
                 clip_border=True,
                 add_ctr_clamp=False,
                 ctr_clamp=32):
        super(BaseBBoxCoder, self).__init__()
        self.means = target_means
        self.stds = target_stds
        self.clip_border = clip_border
        self.add_ctr_clamp = add_ctr_clamp
        self.ctr_clamp = ctr_clamp

    def encode(self, bboxes, gt_bboxes):
        """Get box regression transformation deltas that can be used to
        transform the ``bboxes`` into the ``gt_bboxes``.

        Args:
            bboxes (torch.Tensor): Source boxes, e.g., object proposals.
            gt_bboxes (torch.Tensor): Target of the transformation, e.g.,
                ground-truth boxes.

        Returns:
            torch.Tensor: Box transformation deltas
        """

        assert bboxes.size(0) == gt_bboxes.size(0)
        assert bboxes.size(-1) == gt_bboxes.size(-1) == 4
        encoded_bboxes = bbox2delta(bboxes, gt_bboxes, self.means, self.stds)
        return encoded_bboxes

    def decode(self,
               bboxes,
               pred_bboxes,
               max_shape=None,
               wh_ratio_clip=16 / 1000):
        """Apply transformation `pred_bboxes` to `boxes`.

        Args:
            bboxes (torch.Tensor): Basic boxes. Shape (B, N, 4) or (N, 4)
            pred_bboxes (Tensor): Encoded offsets with respect to each roi.
               Has shape (B, N, num_classes * 4) or (B, N, 4) or
               (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
               when rois is a grid of anchors.Offset encoding follows [1]_.
            max_shape (Sequence[int] or torch.Tensor or Sequence[
               Sequence[int]],optional): Maximum bounds for boxes, specifies
               (H, W, C) or (H, W). If bboxes shape is (B, N, 4), then
               the max_shape should be a Sequence[Sequence[int]]
               and the length of max_shape should also be B.
            wh_ratio_clip (float, optional): The allowed ratio between
                width and height.

        Returns:
            torch.Tensor: Decoded boxes.
        """

        assert pred_bboxes.size(0) == bboxes.size(0)
        if pred_bboxes.ndim == 3:
            assert pred_bboxes.size(1) == bboxes.size(1)

        if pred_bboxes.ndim == 2 and not torch.onnx.is_in_onnx_export():
            # single image decode
            decoded_bboxes = delta2bbox(bboxes, pred_bboxes, self.means,
                                        self.stds, max_shape, wh_ratio_clip,
                                        self.clip_border, self.add_ctr_clamp,
                                        self.ctr_clamp)
        else:
            if pred_bboxes.ndim == 3 and not torch.onnx.is_in_onnx_export():
                warnings.warn(
                    'DeprecationWarning: onnx_delta2bbox is deprecated '
                    'in the case of batch decoding and non-ONNX, '
                    'please use “delta2bbox” instead. In order to improve '
                    'the decoding speed, the batch function will no '
                    'longer be supported. ')
            decoded_bboxes = onnx_delta2bbox(bboxes, pred_bboxes, self.means,
                                             self.stds, max_shape,
                                             wh_ratio_clip, self.clip_border,
                                             self.add_ctr_clamp,
                                             self.ctr_clamp)

        return decoded_bboxes


def bbox2delta(proposals, gt, means=(0., 0., 0., 0.), stds=(1., 1., 1., 1.)):
    """Compute deltas of proposals w.r.t. gt.

    We usually compute the deltas of x, y, w, h of proposals w.r.t ground
    truth bboxes to get regression target.
    This is the inverse function of :func:`delta2bbox`.

    Args:
        proposals (Tensor): Boxes to be transformed, shape (N, ..., 4)
        gt (Tensor): Gt bboxes to be used as base, shape (N, ..., 4)
        means (Sequence[float]): Denormalizing means for delta coordinates
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates

    Returns:
        Tensor: deltas with shape (N, 4), where columns represent dx, dy,
            dw, dh.
    """
    assert proposals.size() == gt.size()

    proposals = proposals.float()
    gt = gt.float()
    px = (proposals[..., 0] + proposals[..., 2]) * 0.5
    py = (proposals[..., 1] + proposals[..., 3]) * 0.5
    pw = proposals[..., 2] - proposals[..., 0]
    ph = proposals[..., 3] - proposals[..., 1]

    gx = (gt[..., 0] + gt[..., 2]) * 0.5
    gy = (gt[..., 1] + gt[..., 3]) * 0.5
    gw = gt[..., 2] - gt[..., 0]
    gh = gt[..., 3] - gt[..., 1]

    dx = (gx - px) / pw
    dy = (gy - py) / ph
    dw = torch.log(gw / pw)
    dh = torch.log(gh / ph)
    deltas = torch.stack([dx, dy, dw, dh], dim=-1)

    means = deltas.new_tensor(means).unsqueeze(0)
    stds = deltas.new_tensor(stds).unsqueeze(0)
    deltas = deltas.sub_(means).div_(stds)

    return deltas


def delta2bbox(rois,
               deltas,
               means=(0., 0., 0., 0.),
               stds=(1., 1., 1., 1.),
               max_shape=None,
               wh_ratio_clip=16 / 1000,
               clip_border=True,
               add_ctr_clamp=False,
               ctr_clamp=32):
    """Apply deltas to shift/scale base boxes.

    Typically the rois are anchor or proposed bounding boxes and the deltas are
    network outputs used to shift/scale those boxes.
    This is the inverse function of :func:`bbox2delta`.

    Args:
        rois (Tensor): Boxes to be transformed. Has shape (N, 4).
        deltas (Tensor): Encoded offsets relative to each roi.
            Has shape (N, num_classes * 4) or (N, 4). Note
            N = num_base_anchors * W * H, when rois is a grid of
            anchors. Offset encoding follows [1]_.
        means (Sequence[float]): Denormalizing means for delta coordinates.
            Default (0., 0., 0., 0.).
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates. Default (1., 1., 1., 1.).
        max_shape (tuple[int, int]): Maximum bounds for boxes, specifies
           (H, W). Default None.
        wh_ratio_clip (float): Maximum aspect ratio for boxes. Default
            16 / 1000.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Default True.
        add_ctr_clamp (bool): Whether to add center clamp. When set to True,
            the center of the prediction bounding box will be clamped to
            avoid being too far away from the center of the anchor.
            Only used by YOLOF. Default False.
        ctr_clamp (int): the maximum pixel shift to clamp. Only used by YOLOF.
            Default 32.

    Returns:
        Tensor: Boxes with shape (N, num_classes * 4) or (N, 4), where 4
           represent tl_x, tl_y, br_x, br_y.

    References:
        .. [1] https://arxiv.org/abs/1311.2524

    Example:
        >>> rois = torch.Tensor([[ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 5.,  5.,  5.,  5.]])
        >>> deltas = torch.Tensor([[  0.,   0.,   0.,   0.],
        >>>                        [  1.,   1.,   1.,   1.],
        >>>                        [  0.,   0.,   2.,  -1.],
        >>>                        [ 0.7, -1.9, -0.5,  0.3]])
        >>> delta2bbox(rois, deltas, max_shape=(32, 32, 3))
        tensor([[0.0000, 0.0000, 1.0000, 1.0000],
                [0.1409, 0.1409, 2.8591, 2.8591],
                [0.0000, 0.3161, 4.1945, 0.6839],
                [5.0000, 5.0000, 5.0000, 5.0000]])
    """
    num_bboxes, num_classes = deltas.size(0), deltas.size(1) // 4
    if num_bboxes == 0:
        return deltas

    deltas = deltas.reshape(-1, 4)

    means = deltas.new_tensor(means).view(1, -1)
    stds = deltas.new_tensor(stds).view(1, -1)
    denorm_deltas = deltas * stds + means

    dxy = denorm_deltas[:, :2]
    dwh = denorm_deltas[:, 2:]

    # Compute width/height of each roi
    rois_ = rois.repeat(1, num_classes).reshape(-1, 4)
    pxy = ((rois_[:, :2] + rois_[:, 2:]) * 0.5)
    pwh = (rois_[:, 2:] - rois_[:, :2])

    dxy_wh = pwh * dxy

    max_ratio = np.abs(np.log(wh_ratio_clip))
    if add_ctr_clamp:
        dxy_wh = torch.clamp(dxy_wh, max=ctr_clamp, min=-ctr_clamp)
        dwh = torch.clamp(dwh, max=max_ratio)
    else:
        dwh = dwh.clamp(min=-max_ratio, max=max_ratio)

    gxy = pxy + dxy_wh
    gwh = pwh * dwh.exp()
    x1y1 = gxy - (gwh * 0.5)
    x2y2 = gxy + (gwh * 0.5)
    bboxes = torch.cat([x1y1, x2y2], dim=-1)
    if clip_border and max_shape is not None:
        bboxes[..., 0::2].clamp_(min=0, max=max_shape[1])
        bboxes[..., 1::2].clamp_(min=0, max=max_shape[0])
    bboxes = bboxes.reshape(num_bboxes, -1)
    return bboxes


def onnx_delta2bbox(rois,
                    deltas,
                    means=(0., 0., 0., 0.),
                    stds=(1., 1., 1., 1.),
                    max_shape=None,
                    wh_ratio_clip=16 / 1000,
                    clip_border=True,
                    add_ctr_clamp=False,
                    ctr_clamp=32):
    """Apply deltas to shift/scale base boxes.

    Typically the rois are anchor or proposed bounding boxes and the deltas are
    network outputs used to shift/scale those boxes.
    This is the inverse function of :func:`bbox2delta`.

    Args:
        rois (Tensor): Boxes to be transformed. Has shape (N, 4) or (B, N, 4)
        deltas (Tensor): Encoded offsets with respect to each roi.
            Has shape (B, N, num_classes * 4) or (B, N, 4) or
            (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
            when rois is a grid of anchors.Offset encoding follows [1]_.
        means (Sequence[float]): Denormalizing means for delta coordinates.
            Default (0., 0., 0., 0.).
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates. Default (1., 1., 1., 1.).
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If rois shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]]
            and the length of max_shape should also be B. Default None.
        wh_ratio_clip (float): Maximum aspect ratio for boxes.
            Default 16 / 1000.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Default True.
        add_ctr_clamp (bool): Whether to add center clamp, when added, the
            predicted box is clamped is its center is too far away from
            the original anchor's center. Only used by YOLOF. Default False.
        ctr_clamp (int): the maximum pixel shift to clamp. Only used by YOLOF.
            Default 32.

    Returns:
        Tensor: Boxes with shape (B, N, num_classes * 4) or (B, N, 4) or
           (N, num_classes * 4) or (N, 4), where 4 represent
           tl_x, tl_y, br_x, br_y.

    References:
        .. [1] https://arxiv.org/abs/1311.2524

    Example:
        >>> rois = torch.Tensor([[ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 0.,  0.,  1.,  1.],
        >>>                      [ 5.,  5.,  5.,  5.]])
        >>> deltas = torch.Tensor([[  0.,   0.,   0.,   0.],
        >>>                        [  1.,   1.,   1.,   1.],
        >>>                        [  0.,   0.,   2.,  -1.],
        >>>                        [ 0.7, -1.9, -0.5,  0.3]])
        >>> delta2bbox(rois, deltas, max_shape=(32, 32, 3))
        tensor([[0.0000, 0.0000, 1.0000, 1.0000],
                [0.1409, 0.1409, 2.8591, 2.8591],
                [0.0000, 0.3161, 4.1945, 0.6839],
                [5.0000, 5.0000, 5.0000, 5.0000]])
    """
    means = deltas.new_tensor(means).view(1,
                                          -1).repeat(1,
                                                     deltas.size(-1) // 4)
    stds = deltas.new_tensor(stds).view(1, -1).repeat(1, deltas.size(-1) // 4)
    denorm_deltas = deltas * stds + means
    dx = denorm_deltas[..., 0::4]
    dy = denorm_deltas[..., 1::4]
    dw = denorm_deltas[..., 2::4]
    dh = denorm_deltas[..., 3::4]

    x1, y1 = rois[..., 0], rois[..., 1]
    x2, y2 = rois[..., 2], rois[..., 3]
    # Compute center of each roi
    px = ((x1 + x2) * 0.5).unsqueeze(-1).expand_as(dx)
    py = ((y1 + y2) * 0.5).unsqueeze(-1).expand_as(dy)
    # Compute width/height of each roi
    pw = (x2 - x1).unsqueeze(-1).expand_as(dw)
    ph = (y2 - y1).unsqueeze(-1).expand_as(dh)

    dx_width = pw * dx
    dy_height = ph * dy

    max_ratio = np.abs(np.log(wh_ratio_clip))
    if add_ctr_clamp:
        dx_width = torch.clamp(dx_width, max=ctr_clamp, min=-ctr_clamp)
        dy_height = torch.clamp(dy_height, max=ctr_clamp, min=-ctr_clamp)
        dw = torch.clamp(dw, max=max_ratio)
        dh = torch.clamp(dh, max=max_ratio)
    else:
        dw = dw.clamp(min=-max_ratio, max=max_ratio)
        dh = dh.clamp(min=-max_ratio, max=max_ratio)
    # Use exp(network energy) to enlarge/shrink each roi
    gw = pw * dw.exp()
    gh = ph * dh.exp()
    # Use network energy to shift the center of each roi
    gx = px + dx_width
    gy = py + dy_height
    # Convert center-xy/width/height to top-left, bottom-right
    x1 = gx - gw * 0.5
    y1 = gy - gh * 0.5
    x2 = gx + gw * 0.5
    y2 = gy + gh * 0.5

    bboxes = torch.stack([x1, y1, x2, y2], dim=-1).view(deltas.size())

    if clip_border and max_shape is not None:
        # clip bboxes with dynamic `min` and `max` for onnx
        if torch.onnx.is_in_onnx_export():
            from mmdet.core.export import dynamic_clip_for_onnx
            x1, y1, x2, y2 = dynamic_clip_for_onnx(x1, y1, x2, y2, max_shape)
            bboxes = torch.stack([x1, y1, x2, y2], dim=-1).view(deltas.size())
            return bboxes
        if not isinstance(max_shape, torch.Tensor):
            max_shape = x1.new_tensor(max_shape)
        max_shape = max_shape[..., :2].type_as(x1)
        if max_shape.ndim == 2:
            assert bboxes.ndim == 3
            assert max_shape.size(0) == bboxes.size(0)

        min_xy = x1.new_tensor(0)
        max_xy = torch.cat(
            [max_shape] * (deltas.size(-1) // 2),
            dim=-1).flip(-1).unsqueeze(-2)
        bboxes = torch.where(bboxes < min_xy, min_xy, bboxes)
        bboxes = torch.where(bboxes > max_xy, max_xy, bboxes)

    return bboxes



import torch
from antgo.framework.helper.models.builder import MODELS, build_model
from antgo.framework.helper.multi_stream_module import MultiSteamModule
from .losses.quality_focal_loss import QualityFocalLoss


@MODELS.register_module()
class DenseTeacher(MultiSteamModule):
    def __init__(self, model: dict, train_cfg=None, test_cfg=None, init_cfg=None):
        # 默认使用teacher模型作为最佳模型
        if test_cfg is None:
            test_cfg = dict()
        test_cfg.update(
            {
                'inference_on': 'teacher'
            }
        )
        super(DenseTeacher, self).__init__(
            dict(
                teacher=build_model(model), 
                student=build_model(model)
            ),
            train_cfg=train_cfg,
            test_cfg=test_cfg,
        )
        self.freeze("teacher")
        self.key = train_cfg.get('key', 0)                                  # 挑选model进入loss前的featuremap list中的哪个作为半监督信号
        self.use_sigmoid = train_cfg.get('use_sigmoid', True)               # 是否将挑选出来的featuremap 使用sigmoid 

        self.label_batch_size = train_cfg.get('label_batch_size', 5)        # 有标签数据量 在一个batch里
        self.unlabel_batch_size = train_cfg.get('unlabel_batch_size', 3)    # 无标签数据量 在一个batch里

        self.semi_ratio = train_cfg.get('semi_ratio',0.5)                   # 挑选出绝对负样本的最大数量
        self.heatmap_n_thr = train_cfg.get('heatmap_n_thr', 0.25)           # heatmap中，小于此值视为绝对负样本，其余位置通过半监督信号监督
        self.semi_loss_w = train_cfg.get('semi_loss_w', 1.0)                # 半监督损失的权重

    def _get_unsup_dense_loss(self, student_heatmap_, teacher_heatmap_):
        # student_heatmap_: N,C,H,W
        # teacher_heatmap_: N,C,H,W
        num_classes = student_heatmap_.shape[1]
        student_heatmap = student_heatmap_.permute(0, 2, 3, 1).reshape(-1, num_classes)
        teacher_heatmap = teacher_heatmap_.permute(0, 2, 3, 1).reshape(-1, num_classes)
        if self.use_sigmoid:
            student_heatmap = student_heatmap.sigmoid()
            teacher_heatmap = teacher_heatmap.sigmoid()

        with torch.no_grad():
            max_vals = torch.max(teacher_heatmap, 1)[0]
            count_num = int(teacher_heatmap.size(0) * self.semi_ratio)
            sorted_vals, sorted_inds = torch.topk(max_vals, teacher_heatmap.size(0))

            for sorted_n in range(count_num):
                if sorted_vals[sorted_n] < self.heatmap_n_thr:
                    count_num = max(1, sorted_n)
                    break

            mask = torch.zeros_like(max_vals)
            mask[sorted_inds[:count_num]] = 1.0
            fg_num = sorted_vals[:count_num].sum()

        loss_heatmap = (
            QualityFocalLoss(
                student_heatmap,
                teacher_heatmap,
                weight=mask,
                reduction="mean",
            )
        )

        return {"loss_heatmap": loss_heatmap}

    def forward_train(self, image, image_meta, **kwargs):
        label_images, unlabel_weak_strong_images = \
            torch.split(image, [self.label_batch_size, self.unlabel_batch_size+self.unlabel_batch_size],dim=0) 
        label_metas = image_meta[:self.label_batch_size]

        unlabel_weak_images=torch.index_select(unlabel_weak_strong_images,dim=0,index=torch.tensor([i for i in range(0,2*self.unlabel_batch_size,2)]))
        unlabel_strong_images=torch.index_select(unlabel_weak_strong_images,dim=0,index=torch.tensor([i for i in range(1,2*self.unlabel_batch_size,2)]))

        unlabel_weak_strong_metas = image_meta[self.label_batch_size:]
        unlabel_weak_metas = [unlabel_weak_strong_metas[i] for i in range(0, 2*self.unlabel_batch_size, 2)]
        unlabel_strong_metas = [unlabel_weak_strong_metas[i] for i in range(1, 2*self.unlabel_batch_size, 2)]
        
        # ignore epoch,iter key in kwargs
        if 'epoch' in kwargs:
            kwargs.pop('epoch')
        if 'iter' in kwargs:
            kwargs.pop('iter')
        
        label_kwargs = {}
        unlabel_weak_kwargs = {}
        unlabel_strong_kwargs = {}
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v_label_kwargs, v_unlabel_weak_strong_kwargs = \
                    torch.split(v, [self.label_batch_size, self.unlabel_batch_size+self.unlabel_batch_size],dim=0) 
            elif isinstance(v, list):
                v_label_kwargs = v[:self.label_batch_size]
                v_unlabel_weak_strong_kwargs = v[self.label_batch_size:]
            else:
                # ignore other
                continue
            
            label_kwargs[k] = v_label_kwargs

            if isinstance(v, torch.Tensor):
                v_unlabel_weak_kwargs=torch.index_select(v_unlabel_weak_strong_kwargs,dim=0,index=torch.tensor([i for i in range(0,2*self.unlabel_batch_size,2)]))
                v_unlabel_strong_kwargs=torch.index_select(v_unlabel_weak_strong_kwargs,dim=0,index=torch.tensor([i for i in range(1,2*self.unlabel_batch_size,2)]))
            else:
                v_unlabel_weak_kwargs = v_unlabel_weak_strong_kwargs[:self.unlabel_batch_size]
                v_unlabel_strong_kwargs = v_unlabel_weak_strong_kwargs[self.unlabel_batch_size:]
            
            unlabel_weak_kwargs[k] = v_unlabel_weak_kwargs
            unlabel_strong_kwargs[k] = v_unlabel_strong_kwargs

        losses = {}
        # 有监督损失
        output_dict = self.student.forward_train(label_images, label_metas, **label_kwargs)
        assert(isinstance(output_dict, dict))
        output_dict = {"labeled_" + k: v for k, v in output_dict.items()}
        losses.update(**output_dict)

        # 无监督损失
        # 应该返回 heatmap
        for k in unlabel_strong_kwargs.keys():
            unlabel_strong_kwargs[k] = None
            unlabel_weak_kwargs[k] = None

        output_unsup_strong = self.student.forward_train(unlabel_strong_images, unlabel_strong_metas, **unlabel_strong_kwargs)
        strong_heatmap = None
        if isinstance(output_unsup_strong, dict):
            strong_heatmap = output_unsup_strong[self.key]
        else:
            strong_heatmap = output_unsup_strong[int(self.key)]

        # 应该返回 heatmap
        self.teacher.eval()
        output_unsup_weak = self.teacher.forward_train(unlabel_weak_images, unlabel_weak_metas, **unlabel_weak_kwargs)
        weak_heatmap = None
        if isinstance(output_unsup_weak, dict):
            weak_heatmap = output_unsup_weak[self.key]
        else:
            weak_heatmap = output_unsup_weak[int(self.key)]

        unsup_loss = self._get_unsup_dense_loss(
            strong_heatmap,
            weak_heatmap
        )

        unsup_loss = {"unlabeled_" + k: v*self.semi_loss_w for k, v in unsup_loss.items()}
        losses.update(**unsup_loss)

        return losses
    
    def simple_test(self, image, image_meta, rescale=True, **kwargs):
        return self.teacher(image, image_meta, rescale, **kwargs)

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs):
        # 同时兼容从base模型和dense模型加载
        if not any(["student" in key or "teacher" in key for key in state_dict.keys()]):
            keys = list(state_dict.keys())
            state_dict.update({"teacher." + k: state_dict[k] for k in keys})
            state_dict.update({"student." + k: state_dict[k] for k in keys})
            for k in keys:
                state_dict.pop(k)

        return super()._load_from_state_dict(
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
        )


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import astor
import os
import os.path as osp
import shutil
import argparse
from six import text_type as _text_type

from .dependency_analyzer import analyze
from .ast_update import update
from .utils import *


def write_file(path, tree):
    code = astor.to_source(tree)
    code = code.replace("(...)", "...")
    code = add_line_continuation_symbol(code)
    f = open(path, "w")
    f.write(code)
    f.close()


def generate_dependencies(folder_path, file_dependencies):
    for name in os.listdir(folder_path):
        current_path = osp.join(folder_path, name)
        if osp.isfile(current_path) and current_path.endswith(".py"):
            if current_path in file_dependencies:
                continue
            analyze(current_path, file_dependencies)
        elif osp.isdir(current_path):
            generate_dependencies(current_path, file_dependencies)


def convert_code(folder_path, new_folder_path, file_dependencies):
    for name in os.listdir(folder_path):
        current_path = osp.join(folder_path, name)
        new_current_path = osp.join(new_folder_path, name)
        if osp.isfile(current_path) and current_path.endswith(".py"):
            print(current_path)
            root = update(current_path, file_dependencies)
            if root is not None:
                write_file(new_current_path, root)
        elif osp.isdir(current_path):
            if not osp.exists(new_current_path):
                os.makedirs(new_current_path)
            convert_code(current_path, new_current_path, file_dependencies)
        elif osp.isfile(current_path) and osp.splitext(current_path)[
                -1] in [".pth", ".pt", ".ckpt"]:
            continue
        elif osp.isfile(current_path) and current_path.endswith(".pyc"):
            continue
        elif osp.isdir(current_path) and current_path == "__pycache__":
            continue
        elif osp.isdir(current_path) and current_path == ".ipynb_checkpoints":
            continue
        else:
            shutil.copyfile(current_path, new_current_path)


def convert_params(params_path):
    import torch
    import paddle
    params = torch.load(params_path, map_location=torch.device('cpu'))
    new_params = dict()
    bn_w_name_list = list()
    for k, v in params.items():
        if k.endswith(".running_mean"):
            new_params[k.replace(".running_mean", "._mean")] = v.detach().numpy(
            )
        elif k.endswith(".running_var"):
            new_params[k.replace(".running_var", "._variance")] = v.detach(
            ).numpy()
            bn_w_name_list.append(k.replace(".running_var", ".weight"))
        else:
            new_params[k] = v.detach().numpy()
    for k, v in new_params.items():
        if len(v.shape) == 2 and k.endswith(
                ".weight") and k not in bn_w_name_list:
            new_params[k] = v.T
    paddle.save(new_params,
                params_path.replace(".pth", ".pdiparams").replace(
                    ".pt", ".pdiparams").replace(".ckpt", ".pdiparams"))


def main(args):
    project_path = args.project_dir
    save_path = args.save_dir
    params_path = args.pretrain_model
    if params_path is not None:
        params_path = osp.abspath(params_path)
        if osp.isdir(params_path):
            for file in os.listdir(params_path):
                if osp.splitext(file)[-1] in [".pth", ".pt", ".ckpt"]:
                    convert_params(osp.join(params_path, file))
        else:
            convert_params(params_path)
    project_path = osp.abspath(project_path)
    file_dependencies = dict()
    sys.path.append(project_path)
    generate_dependencies(project_path, file_dependencies)
    if not osp.exists(save_path):
        os.makedirs(save_path)
    convert_code(project_path, save_path, file_dependencies)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:39
# @File    : repo_normalize.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import re
from urllib.parse import urlparse, parse_qsl, urlsplit
from typing import NamedTuple


class RepoNormalize:
    """
    To normalize the repo.

    Args:
        uri (`str`):
            The uri of the repo.
    """

    def __init__(self, uri: str):
        self._uri = uri
        self._scheme = 'https'
        self._netloc = 'towhee.io'
        self._author = 'towhee'
        self._ref = 'main'
        self._has_ns = True

    def parse_uri(self) -> NamedTuple('ParseResult', [('full_uri', str), ('author', str), ('repo', str), ('ref', str), ('repo_type', str),
                                                      ('norm_repo', str), ('module_name', str), ('class_name', str), ('scheme', str), ('netloc', str),
                                                      ('query', dict)]):
        """
        Parse the uri.

        Returns:
            (`NamedTuple[str, str, str, str, str, str, str, str, str, str, dict]`)
                Return the `full_uri` and its components: `author`, `repo`, `ref`, `repo_type`, `norm_repo`, `module_name`,
                `class_name`, `scheme`, `netloc`, `query`.
        """
        full_uri = self.get_full_uri()
        result = urlparse(full_uri)
        author, repo = result.path.split('/')[1:]
        query = dict(parse_qsl(result.query))
        ref = query.pop('ref')
        norm_repo, module_name, class_name = self.get_name(repo)
        ParseResult = NamedTuple('ParseResult', [('full_uri', str), ('author', str), ('repo', str), ('ref', str), ('repo_type', str),
                                                 ('norm_repo', str), ('module_name', str), ('class_name', str), ('scheme', str), ('netloc', str),
                                                 ('query', dict), ('has_ns', bool)])
        return ParseResult(full_uri, author, repo, ref, result.fragment, norm_repo, module_name, class_name,
                           result.scheme, result.netloc, query, self.has_ns)

    def get_full_uri(self) -> str:
        """
        Get the full uri.

        Returns:
            (`str`)
                The full uri from self._uri.
        """
        result = urlsplit(self._uri)
        path = result.path
        query = dict(parse_qsl(result.query))
        if not result.scheme:
            result = result._replace(scheme=self._scheme)
        if not result.netloc:
            result = result._replace(netloc=self._netloc)
        if 'ref' not in query:
            result = result._replace(query=f'{result.query}&ref={self._ref}')
        if path.endswith(']'):
            path = self.mapping(path)
        if '/' not in path:
            self._has_ns = False
            result = result._replace(path=f'/{self._author}/{path}')
        elif len(path.split('/')) == 2:
            result = result._replace(path=f'/{path}')

        full_uri = result.geturl()
        return full_uri

    def check_uri(self) -> bool:
        """
        Check if the uri matches the format.

        Returns:
            (`bool`)
                Check if passed.

        Raises:
            (`ValueError`)
                Raise error when false.
        """
        result = urlparse(self._uri)
        path = result.path.split('/')
        if len(path) == 1:
            repo = path[0]
        elif len(path) == 2:
            repo = path[1]
        elif len(path) == 3 and path[0] == '':
            repo = path[2]
        else:
            return False
        return self.check_repo(repo)

    def url_valid(self) -> bool:
        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

        if re.match(regex, self._uri) is not None:
            return True
        return False

    @property
    def has_ns(self):
        return self._has_ns

    @staticmethod
    def mapping(path: str) -> str:
        """
        Mapping the path(endswith']'), like repo[framework] to repo-framework.

        Args:
            path (`str`):
                The path to the uri.

        Returns:
            (`str`)
                Check if passed.

        Raises:
            (`ValueError`)
                Raise error when the path does not match the format.
        """
        try:
            if not path.endswith(']'):
                raise ValueError
            repo, framework = path.strip(']').split('[')
            path = f'{repo}-{framework}'
            return path
        except ValueError:
            raise ValueError(f'{path} does not match the \'[/author/]repo-name[framework]\' format!') from ValueError

    @staticmethod
    def check_repo(repo: str) -> bool:
        """
        Check if the repo name matches the format.

        Args:
            repo (`str`):
                Repo name.

        Returns:
            (`bool`)
                Check if matched.

        Raises:
            (`ValueError`)
                Raise error if false.
        """
        repo_list = repo.strip(']').split('[')
        if '_' not in repo and (repo.endswith(']') and len(repo_list) == 2 or len(repo_list) == 1 and not repo.endswith(']')):
            return True
        else:
            raise ValueError(f'repo: {repo} does not match the \'repo-name[framework]\' format!')

    @staticmethod
    def get_name(repo: str) -> NamedTuple('OpResult', [('repo_name', str), ('module_name', str), ('class_name', str)]):
        """
        Get the name for repo: normalized repo name, module name, and class name.

        Args:
            repo (`str`):
                Repo name.

        Returns:
            (`NamedTuple[str, str, str]`)
                Return the name for repo: `repo_name`, `module_name`, `class_name`.
        """
        repo_name = repo.replace('_', '-')
        module_name = repo_name.replace('-', '_')
        class_name = ''.join(x.capitalize() or '_' for x in repo_name.split('-'))
        OpResult = NamedTuple('OpResult', [('repo_name', str), ('module_name', str), ('class_name', str)])
        return OpResult(repo_name, module_name, class_name)

    @staticmethod
    def get_op(repo: str) -> NamedTuple('OpResult', [('repo', str), ('py_file', str), ('yaml_file', str), ('class_name', str)]):
        """
        Get the required name for operators: normalized repo name, python file name, yaml file name and class name.

        Args:
            repo (`str`):
                Repo name.

        Returns:
            (`NamedTuple[str, str, str, str]`)
                Return the required name for operators: `repo`, `py_file`, `yaml_file`, `class_name`.
        """
        repo_name = repo.replace('_', '-')
        file_name = repo_name.replace('-', '_')
        class_name = ''.join(x.capitalize() or '_' for x in repo_name.split('-'))
        OpResult = NamedTuple('OpResult', [('repo', str), ('py_file', str), ('yaml_file', str), ('class_name', str)])
        return OpResult(repo_name, f'{file_name}.py', f'{file_name}.yaml', class_name)

    @staticmethod
    def get_pipeline(repo: str) -> NamedTuple('PipelineResult', [('repo', str), ('yaml_file', str)]):
        """
        Get the required name for pipeline: normalized repo name and yaml file name.

        Args:
            repo (`str`):
                Repo name.

        Returns:
            (`NamedTuple[str, str]`)
                Return the required name for operators: `repo`, `yaml_file`.
        """
        repo_name = repo.replace('_', '-')
        file_name = repo_name.replace('-', '_')
        PipelineResult = NamedTuple('PipelineResult', [('repo', str), ('yaml_file', str)])
        return PipelineResult(repo_name, f'{file_name}.yaml')



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import warnings
from math import inf

import torch.distributed as dist
from torch.nn.modules.batchnorm import _BatchNorm
from torch.utils.data import DataLoader, dataloader

from antgo.framework.helper.fileio import FileClient
from antgo.framework.helper.utils import is_seq_of
from antgo.framework.helper.parallel import is_module_wrapper
from antgo.framework.helper.models.nas.elastic_nn.modules.dynamic_op import *
from antgo.framework.helper.models.nas.elastic_nn.modules.dynamic_layers import *
from antgo.framework.helper.models.nas.utils.common_tools import *
import copy
import torch
import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.runner.hooks.evaluation import EvalHook
from antgo.framework.helper.runner.hooks.hook import HOOKS

@HOOKS.register_module()
class OfaEvalHook(EvalHook):
    """Non-Distributed evaluation hook.

    This hook will regularly perform evaluation in a given interval when
    performing in non-distributed environment.

    Args:
        dataloader (DataLoader): A PyTorch dataloader, whose dataset has
            implemented ``evaluate`` function.
        start (int | None, optional): Evaluation starting epoch. It enables
            evaluation before the training starts if ``start`` <= the resuming
            epoch. If None, whether to evaluate is merely decided by
            ``interval``. Default: None.
        interval (int): Evaluation interval. Default: 1.
        by_epoch (bool): Determine perform evaluation by epoch or by iteration.
            If set to True, it will perform by epoch. Otherwise, by iteration.
            Default: True.
        save_best (str, optional): If a metric is specified, it would measure
            the best checkpoint during evaluation. The information about best
            checkpoint would be saved in ``runner.meta['hook_msgs']`` to keep
            best score value and best checkpoint path, which will be also
            loaded when resume checkpoint. Options are the evaluation metrics
            on the test dataset. e.g., ``bbox_mAP``, ``segm_mAP`` for bbox
            detection and instance segmentation. ``AR@100`` for proposal
            recall. If ``save_best`` is ``auto``, the first key of the returned
            ``OrderedDict`` result will be used. Default: None.
        rule (str | None, optional): Comparison rule for best score. If set to
            None, it will infer a reasonable rule. Keys such as 'acc', 'top'
            .etc will be inferred by 'greater' rule. Keys contain 'loss' will
            be inferred by 'less' rule. Options are 'greater', 'less', None.
            Default: None.
        test_fn (callable, optional): test a model with samples from a
            dataloader, and return the test results. If ``None``, the default
            test function ``mmcv.engine.single_gpu_test`` will be used.
            (default: ``None``)
        greater_keys (List[str] | None, optional): Metric keys that will be
            inferred by 'greater' comparison rule. If ``None``,
            _default_greater_keys will be used. (default: ``None``)
        less_keys (List[str] | None, optional): Metric keys that will be
            inferred by 'less' comparison rule. If ``None``, _default_less_keys
            will be used. (default: ``None``)
        out_dir (str, optional): The root directory to save checkpoints. If not
            specified, `runner.work_dir` will be used by default. If specified,
            the `out_dir` will be the concatenation of `out_dir` and the last
            level directory of `runner.work_dir`.
            `New in version 1.3.16.`
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details. Default: None.
            `New in version 1.3.16.`
        **eval_kwargs: Evaluation arguments fed into the evaluate function of
            the dataset.

    Note:
        If new arguments are added for EvalHook, tools/test.py,
        tools/eval_metric.py may be affected.
    """

    # Since the key for determine greater or less is related to the downstream
    # tasks, downstream repos may need to overwrite the following inner
    # variable accordingly.

    rule_map = {'greater': lambda x, y: x > y, 'less': lambda x, y: x < y}
    init_value_map = {'greater': -inf, 'less': inf}
    _default_greater_keys = [
        'acc', 'top', 'AR@', 'auc', 'precision', 'mAP', 'mDice', 'mIoU',
        'mAcc', 'aAcc'
    ]
    _default_less_keys = ['loss']

    def __init__(self,
                 dataloader,
                 start=None,
                 interval=1,
                 by_epoch=True,
                 save_best=None,
                 rule=None,
                 test_fn=None,
                 greater_keys=None,
                 less_keys=None,
                 out_dir=None,
                 file_client_args=None,
                 metric=None,
                 image_size_list=None,
                 ks_list=None,
                 expand_ratio_list=None,
                 depth_list=None,
                 width_mult_list=None,
                 **eval_kwargs):
        super().__init__(                 
            dataloader, 
            start, 
            interval, 
            by_epoch, 
            save_best, 
            rule, 
            test_fn, 
            greater_keys, 
            less_keys, 
            out_dir, 
            file_client_args, 
            metric, 
            **eval_kwargs
        )
        self.image_size_list = image_size_list
        self.ks_list = ks_list
        self.expand_ratio_list = expand_ratio_list
        self.depth_list = depth_list
        self.width_mult_list = width_mult_list

    def reset_running_statistics(self, model):
        bn_mean = {}
        bn_var = {}

        forward_model = copy.deepcopy(model)
        forward_model.eval()
        for name, m in forward_model.named_modules():
            if isinstance(m, nn.BatchNorm2d):
                bn_mean[name] = AverageMeter()
                bn_var[name] = AverageMeter()

                def new_forward(bn, mean_est, var_est):
                    def lambda_forward(x):
                        batch_mean = (
                            x.mean(0, keepdim=True)
                            .mean(2, keepdim=True)
                            .mean(3, keepdim=True)
                        )  # 1, C, 1, 1
                        batch_var = (x - batch_mean) * (x - batch_mean)
                        batch_var = (
                            batch_var.mean(0, keepdim=True)
                            .mean(2, keepdim=True)
                            .mean(3, keepdim=True)
                        )

                        batch_mean = torch.squeeze(batch_mean)
                        batch_var = torch.squeeze(batch_var)

                        mean_est.update(batch_mean.data, x.size(0))
                        var_est.update(batch_var.data, x.size(0))

                        # bn forward using calculated mean & var
                        _feature_dim = batch_mean.size(0)
                        return F.batch_norm(
                            x,
                            batch_mean,
                            batch_var,
                            bn.weight[:_feature_dim],
                            bn.bias[:_feature_dim],
                            False,
                            0.0,
                            bn.eps,
                        )

                    return lambda_forward

                m.forward = new_forward(m, bn_mean[name], bn_var[name])

        if len(bn_mean) == 0:
            # skip if there is no batch normalization layers in the network
            return

        with torch.no_grad():
            DynamicBatchNorm2d.SET_RUNNING_STATISTICS = True
            for data in self.dataloader:
                if type(data) == list or type(data) == tuple:
                    forward_model(*data, return_loss=False)
                else:
                    data.update({
                        'return_loss': False
                    })
                    forward_model(**data)
            DynamicBatchNorm2d.SET_RUNNING_STATISTICS = False

        for name, m in model.named_modules():
            if name in bn_mean and bn_mean[name].count > 0:
                feature_dim = bn_mean[name].avg.size(0)
                assert isinstance(m, nn.BatchNorm2d)
                m.running_mean.data[:feature_dim].copy_(bn_mean[name].avg)
                m.running_var.data[:feature_dim].copy_(bn_var[name].avg)

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        dynamic_net = runner.model
        if is_module_wrapper(dynamic_net):
            dynamic_net = dynamic_net.module
        # 
        image_size_list = self.image_size_list
        if image_size_list is None:
            image_size_list = val2list(self.dataloader.dataset.get_active_target_size(), 1)
        ks_list = self.ks_list
        if ks_list is None:
            ks_list = dynamic_net.ks_list
        expand_ratio_list = self.expand_ratio_list
        if expand_ratio_list is None:
            expand_ratio_list = dynamic_net.expand_ratio_list
        
        depth_list = self.depth_list
        if depth_list is None:
            depth_list = dynamic_net.depth_list
        
        width_mult_list = self.width_mult_list
        if width_mult_list is None:
            if "width_mult_list" in dynamic_net.__dict__:
                width_mult_list = list(range(len(dynamic_net.width_mult_list)))
            else:
                width_mult_list = [0]

        subnet_settings = []
        for d in depth_list:
            for e in expand_ratio_list:
                for k in ks_list:
                    for w in width_mult_list:
                        for img_size in image_size_list:
                            subnet_settings.append(
                                [
                                    {
                                        "image_size": img_size,
                                        "d": d,
                                        "e": e,
                                        "ks": k,
                                        "w": w,
                                    },
                                    "R%s-D%s-E%s-K%s-W%s" % (img_size, d, e, k, w),
                                ]
                            )

        for setting, name in subnet_settings:
            self.dataloader.dataset.set_active_target_size(
                setting.pop("image_size")
            )
            # 基于配置，设置当前激活
            dynamic_net.set_active_subnet(**setting)

            # 重新统计当前激活子网络BN
            self.reset_running_statistics(runner.model)

            # 计算网络输出结果
            results = self.test_fn(runner.model, self.dataloader)
            runner.log_buffer.output['eval_iter_num'] = len(self.dataloader)

            # 评估指标
            key_score = self.evaluate(runner, results)

            # 保存模型
            if self.save_best and key_score:
                self._save_ckpt(runner, key_score)


@HOOKS.register_module()
class OfaDistEvalHook(OfaEvalHook):
    """Distributed evaluation hook.

    This hook will regularly perform evaluation in a given interval when
    performing in distributed environment.

    Args:
        dataloader (DataLoader): A PyTorch dataloader, whose dataset has
            implemented ``evaluate`` function.
        start (int | None, optional): Evaluation starting epoch. It enables
            evaluation before the training starts if ``start`` <= the resuming
            epoch. If None, whether to evaluate is merely decided by
            ``interval``. Default: None.
        interval (int): Evaluation interval. Default: 1.
        by_epoch (bool): Determine perform evaluation by epoch or by iteration.
            If set to True, it will perform by epoch. Otherwise, by iteration.
            default: True.
        save_best (str, optional): If a metric is specified, it would measure
            the best checkpoint during evaluation. The information about best
            checkpoint would be saved in ``runner.meta['hook_msgs']`` to keep
            best score value and best checkpoint path, which will be also
            loaded when resume checkpoint. Options are the evaluation metrics
            on the test dataset. e.g., ``bbox_mAP``, ``segm_mAP`` for bbox
            detection and instance segmentation. ``AR@100`` for proposal
            recall. If ``save_best`` is ``auto``, the first key of the returned
            ``OrderedDict`` result will be used. Default: None.
        rule (str | None, optional): Comparison rule for best score. If set to
            None, it will infer a reasonable rule. Keys such as 'acc', 'top'
            .etc will be inferred by 'greater' rule. Keys contain 'loss' will
            be inferred by 'less' rule. Options are 'greater', 'less', None.
            Default: None.
        test_fn (callable, optional): test a model with samples from a
            dataloader in a multi-gpu manner, and return the test results. If
            ``None``, the default test function ``mmcv.engine.multi_gpu_test``
            will be used. (default: ``None``)
        tmpdir (str | None): Temporary directory to save the results of all
            processes. Default: None.
        gpu_collect (bool): Whether to use gpu or cpu to collect results.
            Default: False.
        broadcast_bn_buffer (bool): Whether to broadcast the
            buffer(running_mean and running_var) of rank 0 to other rank
            before evaluation. Default: True.
        out_dir (str, optional): The root directory to save checkpoints. If not
            specified, `runner.work_dir` will be used by default. If specified,
            the `out_dir` will be the concatenation of `out_dir` and the last
            level directory of `runner.work_dir`.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmcv.fileio.FileClient` for details. Default: None.
        **eval_kwargs: Evaluation arguments fed into the evaluate function of
            the dataset.
    """

    def __init__(self,
                 dataloader,
                 start=None,
                 interval=1,
                 by_epoch=True,
                 save_best=None,
                 rule=None,
                 test_fn=None,
                 greater_keys=None,
                 less_keys=None,
                 broadcast_bn_buffer=True,
                 tmpdir=None,
                 gpu_collect=False,
                 out_dir=None,
                 file_client_args=None,
                 metric=None,
                 image_size_list=None,
                 ks_list=None,
                 expand_ratio_list=None,
                 depth_list=None,
                 width_mult_list=None,                 
                 **eval_kwargs):
        super().__init__(
                 dataloader, 
                 start, 
                 interval, 
                 by_epoch, 
                 save_best,
                 rule, 
                 test_fn, 
                 greater_keys, 
                 less_keys, 
                 broadcast_bn_buffer, 
                 tmpdir, 
                 gpu_collect, 
                 out_dir, 
                 file_client_args, 
                 metric, 
                 image_size_list,
                 ks_list,
                 expand_ratio_list,
                 depth_list,
                 width_mult_list,
                 **eval_kwargs            
        )

        self.broadcast_bn_buffer = broadcast_bn_buffer
        self.tmpdir = tmpdir
        self.gpu_collect = gpu_collect

    def reset_running_statistics(self, model):
        bn_mean = {}
        bn_var = {}

        forward_model = copy.deepcopy(model)
        for name, m in forward_model.named_modules():
            if isinstance(m, nn.BatchNorm2d):
                def new_forward(bn):
                    def lambda_forward(x):
                        batch_mean = (
                            x.mean(0, keepdim=True)
                            .mean(2, keepdim=True)
                            .mean(3, keepdim=True)
                        )  # 1, C, 1, 1
                        batch_var = (x - batch_mean) * (x - batch_mean)
                        batch_var = (
                            batch_var.mean(0, keepdim=True)
                            .mean(2, keepdim=True)
                            .mean(3, keepdim=True)
                        )

                        batch_mean = torch.squeeze(batch_mean)
                        batch_var = torch.squeeze(batch_var)

                        # 需要所有节点同步平均
                        dist.all_reduce(batch_mean, op=dist.reduce_op.AVG, group=0)
                        dist.all_reduce(batch_var, op=dist.reduce_op.AVG, group=0)

                        # bn forward using calculated mean & var
                        _feature_dim = batch_mean.size(0)
                        return F.batch_norm(
                            x,
                            batch_mean,
                            batch_var,
                            bn.weight[:_feature_dim],
                            bn.bias[:_feature_dim],
                            False,
                            0.0,
                            bn.eps,
                        )

                    return lambda_forward

                m.forward = new_forward(m)

        if len(bn_mean) == 0:
            # skip if there is no batch normalization layers in the network
            return

        with torch.no_grad():
            DynamicBatchNorm2d.SET_RUNNING_STATISTICS = True
            for images, labels in self.dataloader:
                images = images.to(forward_model.parameters().__next__().device)
                forward_model(images)
            DynamicBatchNorm2d.SET_RUNNING_STATISTICS = False

        for name, m in model.named_modules():
            if name in bn_mean and bn_mean[name].count > 0:
                feature_dim = bn_mean[name].avg.size(0)
                assert isinstance(m, nn.BatchNorm2d)
                m.running_mean.data[:feature_dim].copy_(bn_mean[name].avg)
                m.running_var.data[:feature_dim].copy_(bn_var[name].avg)

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        # Synchronization of BatchNorm's buffer (running_mean
        # and running_var) is not supported in the DDP of pytorch,
        # which may cause the inconsistent performance of models in
        # different ranks, so we broadcast BatchNorm's buffers
        # of rank 0 to other ranks to avoid this.
        """perform evaluation and save ckpt."""
        dynamic_net = runner.model
        if is_module_wrapper(dynamic_net):
            dynamic_net = dynamic_net.module
        # 
        image_size_list = self.image_size_list
        if image_size_list is None:
            image_size_list = val2list(self.dataloader.dataset.get_active_target_size(), 1)
        ks_list = self.ks_list
        if ks_list is None:
            ks_list = dynamic_net.ks_list
        expand_ratio_list = self.expand_ratio_list
        if expand_ratio_list is None:
            expand_ratio_list = dynamic_net.expand_ratio_list
        
        depth_list = self.depth_list
        if depth_list is None:
            depth_list = dynamic_net.depth_list
        
        width_mult_list = self.width_mult_list
        if width_mult_list is None:
            if "width_mult_list" in dynamic_net.__dict__:
                width_mult_list = list(range(len(dynamic_net.width_mult_list)))
            else:
                width_mult_list = [0]

        subnet_settings = []
        for d in depth_list:
            for e in expand_ratio_list:
                for k in ks_list:
                    for w in width_mult_list:
                        for img_size in image_size_list:
                            subnet_settings.append(
                                [
                                    {
                                        "image_size": img_size,
                                        "d": d,
                                        "e": e,
                                        "ks": k,
                                        "w": w,
                                    },
                                    "R%s-D%s-E%s-K%s-W%s" % (img_size, d, e, k, w),
                                ]
                            )

        for setting, name in subnet_settings:
            self.dataloader.dataset.set_active_target_size(
                setting.pop("image_size")
            )
            # 基于配置，设置当前激活
            dynamic_net.set_active_subnet(**setting)

            # 重新统计当前激活子网络BN
            self.reset_running_statistics(runner.model)

            # 计算网络输出结果
            results = self.test_fn(runner.model, self.dataloader)
            runner.log_buffer.output['eval_iter_num'] = len(self.dataloader)

            # 计算网络输出结果
            tmpdir = self.tmpdir
            if tmpdir is None:
                tmpdir = osp.join(runner.work_dir, '.eval_hook')

            # 所有节点都需要运行
            results = self.test_fn(
                runner.model,
                self.dataloader,
                tmpdir=tmpdir,
                gpu_collect=False)

            if runner.rank == 0:
                # 仅在master节点进行统计模型分数
                print('\n')
                runner.log_buffer.output['eval_iter_num'] = len(self.dataloader)
                key_score = self.evaluate(runner, results)
                # the key_score may be `None` so it needs to skip the action to
                # save the best checkpoint
                if self.save_best and key_score:
                    self._save_ckpt(runner, key_score)



import torch
import torch.nn as nn
from antgo.framework.helper.cnn.bricks import *
from antgo.framework.helper.models.builder import HEADS, MODELS, build_loss
from antgo.framework.helper.runner import BaseModule
import torch.nn.functional as F


@HEADS.register_module()
class SimpleHead(BaseModule):
    def __init__(self, 
                 in_channels=160, 
                 channels=32, 
                 out_channels=2, 
                 loss_decode=dict(
                     type='CrossEntropyLoss',
                     use_sigmoid=False,
                     loss_weight=1.0),
                 align_corners=False,
                 ignore_index=None,
                 **kwargs):
        super().__init__(init_cfg=kwargs.get('init_cfg', None))
        self.in_channels = in_channels
        self.channels = channels
        self.out_channels = out_channels

        self.loss_decode = build_loss(loss_decode)
        self.ignore_index = ignore_index

        self.bottleneck = ConvModule(
            self.in_channels,
            self.channels,
            3,
            padding=1,
            conv_cfg=None,
            norm_cfg=dict(type='BN'),
            act_cfg=dict(type='ReLU'))
        self.conv_seg = nn.Conv2d(channels, self.out_channels, kernel_size=1)
        self.align_corners = align_corners

    def cls_seg(self, feat):
        """Classify each pixel."""
        # if self.dropout is not None:
        #     feat = self.dropout(feat)
        output = self.conv_seg(feat)
        return output

    def forward(self, inputs):
        """Forward function."""
        if isinstance(inputs, list) or isinstance(inputs, tuple):
            inputs = inputs[-1]

        feats = self.bottleneck(inputs)
        output = self.cls_seg(feats)
        return output

    def loss(self, seg_logits, seg_labels):
        loss_dict = dict()
        seg_logits = F.interpolate(seg_logits, seg_labels.shape[2:], mode='bilinear', align_corners=self.align_corners)
        seg_labels = seg_labels.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        for loss_decode in losses_decode:
            loss_dict['loss_seg'] = loss_decode(
                seg_logits,
                seg_labels,
                ignore_index=self.ignore_index)

        return loss_dict
    
    def simple_test(self, x):
        x = self.forward(x)
        x = torch.softmax(x, 1)
        return x


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 14:31
# @File    : dispatcher.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.pipeline.engine.factory import ops


class DispatcherMixin:
    """
    Mixin for call dispatcher for data collection

    >>> @register(name='add_1')
    ... def add_1(x):
    ...     return x+1

    >>> dc = DataCollection.range(5).stream()
    >>> dc.add_1['a','b','c']() #doctest: +ELLIPSIS
    <map object at ...>
    """

    def resolve(self, path, index, *arg, **kws):
        return getattr(ops, path)[index](*arg, **kws)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/13 12:30
# @File    : utils.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import platform
import subprocess
import time
from pathlib import Path
import requests
import torch


def select_device(device=''):
    # device = 'cpu' or '0' or '0,1,2,3'
    cpu = device.lower() == 'cpu'
    if cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # force torch.cuda.is_available() = False
    elif device:  # non-cpu device requested
        os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
        assert torch.cuda.is_available(), f'CUDA unavailable, invalid device {device} requested'  # check availability

    cuda = not cpu and torch.cuda.is_available()
    return torch.device('cuda:0' if cuda else 'cpu')


def torch_model_download(file, repo='jianzfb/antgo', file_type='.pt'):
    # Attempt file download if does not exist
    if not file.endswith(file_type):
        file = f'{file}{file_type}'

    file = file.lower()
    model_file = os.path.join(os.environ['HOME'], '.antgo', 'models', file)
    model_file = Path(model_file)
    if model_file.exists():
        return model_file

    response = requests.get(f'https://api.github.com/repos/{repo}/releases/latest').json()  # github api
    assets = [x['name'] for x in response['assets']]  # release assets
    tag = response['tag_name']  # i.e. 'v1.0'

    name = model_file.name
    if name in assets:
        msg = f'{model_file} missing, try downloading from https://github.com/{repo}/releases/'
        try:  # GitHub
            url = f'https://github.com/{repo}/releases/download/{tag}/{name}'
            print(f'Downloading {url} to {model_file}...')
            torch.hub.download_url_to_file(url, str(model_file))
            assert model_file.exists() and model_file.stat().st_size > 1E6  # check
        except Exception as e:  # GCP
            print(f'Download {name} error.')
        finally:
            if not model_file.exists() or model_file.stat().st_size < 1E6:  # check
                model_file.unlink(missing_ok=True)  # remove partial downloads
                print(f'ERROR: Download failure: {msg}')
            print('')
            return model_file


from .iou_calculators import BboxOverlaps2D, bbox_overlaps
from .coder import (BaseBBoxCoder, DeltaXYWHBBoxCoder, DistancePointBBoxCoder,
                    PseudoBBoxCoder, TBLRBBoxCoder)
from .assigners import *
from .samplers import *
from .transforms import (bbox2distance, bbox2result, bbox2roi,
                         bbox_cxcywh_to_xyxy, bbox_flip, bbox_mapping,
                         bbox_mapping_back, bbox_rescale, bbox_xyxy_to_cxcywh,
                         distance2bbox, roi2bbox)

__all__ = [
    'bbox_overlaps', 'BboxOverlaps2D', 'DeltaXYWHBBoxCoder', 
    'DistancePointBBoxCoder', 'PseudoBBoxCoder', 'TBLRBBoxCoder',
    'bbox2distance', 'bbox2result', 'bbox2roi',
    'bbox_cxcywh_to_xyxy', 'bbox_flip', 'bbox_mapping',
    'bbox_mapping_back', 'bbox_rescale', 'bbox_xyxy_to_cxcywh',
    'distance2bbox', 'roi2bbox'
]


import os
from antgo.pipeline.extent import op

for op_file_name in os.listdir(os.path.dirname(__file__)):
    if op_file_name[0] == '.':
        continue
    if op_file_name.startswith('_'):
        continue

    op.load(op_file_name, os.path.dirname(__file__))


# -*- coding: UTF-8 -*-
# File: mnist.py
# Author: jian(jian@mltalker.com)
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
import os
import gzip
import random
import numpy as np
from antgo.dataflow.dataset.dataset import *
import time
import copy

__all__ = ['Mnist']
MINIST_URL = 'http://yann.lecun.com/exdb/mnist/'

def _read32(bytestream):
  dt = np.dtype(np.uint32).newbyteorder('>')
  return np.frombuffer(bytestream.read(4), dtype=dt)[0]


def extract_images(filename):
  """Extract the images into a 4D uint8 numpy array [index, y, x, depth]."""
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2051:
      raise ValueError(
         'Invalid magic number %d in MNIST image file: %s' %
         (magic, filename))
    num_images = _read32(bytestream)
    rows = _read32(bytestream)
    cols = _read32(bytestream)
    buf = bytestream.read(rows * cols * num_images)
    data = np.frombuffer(buf, dtype=np.uint8)
    data = data.reshape(num_images, rows, cols, 1)
    return data

def extract_labels(filename):
  """Extract the labels into a 1D uint8 numpy array [index]."""
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2049:
        raise ValueError(
          'Invalid magic number %d in MNIST label file: %s' %
          (magic, filename))
    num_items = _read32(bytestream)
    buf = bytestream.read(num_items)
    labels = np.frombuffer(buf, dtype=np.uint8)
    return labels

class DataSet(object):
  def __init__(self, images, labels, fake_data=False):
    """Construct a DataSet. """
    assert images.shape[0] == labels.shape[0], (
        'images.shape: %s labels.shape: %s' % (images.shape,
                                               labels.shape))
    self._num_examples = images.shape[0]

    # Convert shape from [num examples, rows, columns, depth]
    # to [num examples, rows*columns] (assuming depth == 1)
    assert images.shape[3] == 1
    images = images.reshape(images.shape[0],
                            images.shape[1] * images.shape[2])
    # Convert from [0, 255] -> [0.0, 1.0].
    images = images.astype(np.float32)
    images = np.multiply(images, 1.0 / 255.0)
    self._images = images
    self._labels = labels

  @property
  def images(self):
    return self._images

  @property
  def labels(self):
    return self._labels

  @property
  def num_examples(self):
    return self._num_examples


class Mnist(Dataset):
  """
  Return [image, label],
      image is 28x28 in the range [0,1]
  """
  def __init__(self, train_or_test, dir=None, ext_params=None):
    """
    Args:
        train_or_test: string either 'train' or 'test'
    """
    super(Mnist,self).__init__(train_or_test, dir, ext_params=ext_params)
    self.train_or_test = train_or_test

    if self.train_or_test == "train":
      self.download(self.dir, file_names=['train-images-idx3-ubyte.gz'], default_url=MINIST_URL)
      self.train_images = extract_images(os.path.join(self.dir, 'train-images-idx3-ubyte.gz'))

      self.download(self.dir, file_names=['train-labels-idx1-ubyte.gz'], default_url=MINIST_URL)
      self.train_labels = extract_labels(os.path.join(self.dir, 'train-labels-idx1-ubyte.gz'))
      self.train = DataSet(self.train_images, self.train_labels)

      self.ids = [i for i in range(self.train.num_examples)]
    else:
      self.download(self.dir, file_names=['t10k-images-idx3-ubyte.gz'], default_url=MINIST_URL)
      test_images = extract_images(os.path.join(self.dir,'t10k-images-idx3-ubyte.gz'))

      self.download(self.dir, file_names=['t10k-labels-idx1-ubyte.gz'], default_url=MINIST_URL)
      test_labels = extract_labels(os.path.join(self.dir, 't10k-labels-idx1-ubyte.gz'))
      self.test = DataSet(test_images, test_labels)

      self.ids = [i for i in range(self.test.num_examples)]

  def split(self, split_params={}, split_method='holdout'):
      assert(self.train_or_test == 'train')
      assert(split_method == 'holdout')
      val_dataset = Mnist('test', self.dir, self.ext_params)
      return self, val_dataset

  @property
  def size(self):
    return len(self.ids)

  def at(self, id):
    if self.train_or_test == 'train':
      img = self.train.images[id].reshape((28, 28,1))
      img = np.concatenate([img,img,img],-1)
      img = img.astype(np.uint8)
      label = self.train.labels[id]

      return img, label
    else:
      img = self.test.images[id].reshape((28, 28,1))
      img = np.concatenate([img, img, img], -1)
      img = img.astype(np.uint8)
      label = self.test.labels[id]

      return img, label

# mnist = Mnist('val', "/root/workspace/dataset/temp_dataset")
# for i in range(mnist.size):
#   data = mnist.sample(i)
#   print(i)


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:03
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from .hyperparameter import HyperParameter
from .hyperparameter import param_scope, reads, writes, all_params
from .hyperparameter import auto_param, set_auto_param_callback
from .hyperparameter import dynamic_dispatch

__all__ = [
    'HyperParameter', 'dynamic_dispatch', 'param_scope', 'reads', 'writes',
    'all_params', 'auto_param', 'set_auto_param_callback'
]



# encoding=utf-8
# @Time    : 17-3-3
# @File    : main.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import sys
import os
from antgo.utils.utils import *
from antgo.utils.args import *
from antgo.ant.utils import *
from antgo.ant.client import get_client, launch_server
from antgo.command import *
from antgo.framework.helper.tools.util import *
from antgo.help import *
from antgo import config
from jinja2 import Environment, FileSystemLoader
import json
from antgo import tools
from antgo.script import *
import yaml


# 需要使用python3
assert(int(sys.version[0]) >= 3)


#############################################
#######   antgo parameters            #######
#############################################
DEFINE_string('project', '', 'set project name')
DEFINE_string('git', None, '')
DEFINE_string('branch', None, '')
DEFINE_string('commit', None, '')
DEFINE_string('image', '', '')      # 镜像
DEFINE_int('cpu', 0, 'set cpu number')          # cpu 数
DEFINE_int('gpu', 0, 'set gpu number')          # gpu 数
DEFINE_int('memory', 0, 'set memory size (M)')  # 内存大小（单位M）
DEFINE_string('name', None, '')     # 名字
DEFINE_string('type', None, '')     # 类型
DEFINE_indicator('cloud', True, '') # 指定云端运行
DEFINE_string("address", None, "")
DEFINE_indicator("auto", True, '')  # 是否项目自动优化
DEFINE_indicator("finetune", True, '')  # 是否启用finetune模式
DEFINE_string('id', None, '')
DEFINE_int('port', 0, 'set port')
DEFINE_choices('stage', 'supervised', ['supervised', 'semi-supervised', 'distillation', 'activelearning', 'label'], '')

############## submitter ###################
DEFINE_indicator('ssh', True, '')     # ssh 提交
DEFINE_indicator('k8s', True, '')     # k8s 提交
DEFINE_indicator('local', True, '')   # 本地提交

############## global config ###############
DEFINE_string('token', None, '')
############## project config ##############
DEFINE_string('semi', "", 'set semi supervised method')
DEFINE_string("distillation", "", "set distillation method")
DEFINE_string("activelearning", "", "set activelearning method")
DEFINE_string("ensemble", "", "set ensemble method")

############## tool config ##############
DEFINE_string("src", "./", "set src folder/file")
DEFINE_string("tgt", "./", "set tgt folder/file")
DEFINE_int("frame-rate", 30, "video frame rate")
DEFINE_string("prefix", None, "filter by prefixe")
DEFINE_string("suffix", None, "filter by suffix")
DEFINE_string("ext", None, "filter by ext")
DEFINE_string("white-users", None, "name:password,name:password")
DEFINE_string("tags", None, "tag info")
DEFINE_string("no-tags", None, "tag info")
DEFINE_indicator("feedback", True, "")
DEFINE_indicator("user-input", True, "")
DEFINE_int('num', 0, "number")
DEFINE_indicator("to", True, "")
DEFINE_indicator("from", True, "")
DEFINE_indicator("extra", True, "load extra package(mano,smpl,...)")
DEFINE_indicator("shuffle", True, "")
DEFINE_int("max-size", 0, "")
DEFINE_float("ext-ratio", 0.35, "")
DEFINE_int("ext-size", 0, "")
DEFINE_indicator("ignore-incomplete", True, "")

#############################################
DEFINE_nn_args()

action_level_1 = ['train', 'eval', 'export', 'config', 'submitter', 'server', 'activelearning']
action_level_2 = ['add', 'del', 'create', 'register','update', 'show', 'get', 'tool', 'share']


def main():
  # 显示antgo logo
  main_logo()
  
  # 备份脚本参数
  sys_argv_cp = sys.argv
  
  # 解析参数
  action_name = sys.argv[1]
  sub_action_name = None
  if action_name in action_level_1:
    sys.argv = [sys.argv[0]] + sys.argv[2:]
  else:
    sub_action_name = sys.argv[2]
    sys.argv = [sys.argv[0]] + sys.argv[3:]
  args = parse_args()

  ######################################### 配置文件操作 ################################################
  # 检查配置文件是否存在
  if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')):
    # 使用指定配置文件更新
    if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'antgo')):
      os.makedirs(os.path.join(os.environ['HOME'], '.config', 'antgo'))

    config_data = {'FACTORY': './.factory', 'USER_TOKEN': ''}
    env = Environment(loader=FileSystemLoader('/'.join(os.path.realpath(__file__).split('/')[0:-1])))
    config_template = env.get_template('config.xml')
    config_content = config_template.render(**config_data)

    with open(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'), 'w') as fp:
      fp.write(config_content)
    logging.warn('Using default config file.')

  # 配置操作
  if action_name == 'config':
    config_data = {'FACTORY': '', 'USER_TOKEN': ''}
    # 读取现有数值
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    config.AntConfig.parse_xml(config_xml)
    config_data['FACTORY'] = getattr(config.AntConfig, 'factory', '')
    config_data['USER_TOKEN'] = getattr(config.AntConfig, 'token', '')

    if args.root is not None:
      config_data['FACTORY'] = args.root
    if args.token is not None:
      config_data['USER_TOKEN'] = args.token

    env = Environment(loader=FileSystemLoader('/'.join(os.path.realpath(__file__).split('/')[0:-1])))
    config_template = env.get_template('config.xml')
    config_content = config_template.render(**config_data)

    with open(os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml'), 'w') as fp:
      fp.write(config_content)

    logging.info('Update config file.')
    return

  # 解析配置文件
  config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
  config.AntConfig.parse_xml(config_xml)

  if not os.path.exists(config.AntConfig.factory):
    os.makedirs(config.AntConfig.factory)
  if not os.path.exists(config.AntConfig.data_factory):
    os.makedirs(config.AntConfig.data_factory)
  if not os.path.exists(config.AntConfig.task_factory):
    os.makedirs(config.AntConfig.task_factory)

  ######################################### 支持扩展 ###############################################
  if args.extra and not os.path.exists('extra'):
    logging.info('download extra package')
    os.system('wget http://image.mltalker.com/extra.tar; tar -xf extra.tar; cd extra/manopth; python3 setup.py install')
  
  if args.ext_module != '':
    logging.info('import extent module')
    load_extmodule(args.ext_module)

  ##################################### 支持任务提交脚本配置  ###########################################
  if action_name == 'submitter':
    has_config_file = not (args.config == '' or args.config == 'config.py')
    has_config_file = has_config_file and os.path.exists(args.config)
    
    if has_config_file:
      if not (args.ssh or args.local or args.k8s):
        # 当前配置脚本为用户定制化提交脚本（在提交任务时，优先使用此脚本提交）
        # 任务提交脚本，参数必须如下格式:
        # 镜像名字, 启动命令, GPU 数, CPU 数, 内存大小
        # image_name, launch_argv, gpu_num, cpu_num, memory_size
        inner_folder = os.path.join(os.environ['HOME'], '.config', 'antgo')
        with open(args.config, encoding='utf-8', mode='r') as fp:
          config_content = yaml.safe_load(fp)

        # 将配置脚本目录写入内部目录
        code_folder = config_content['folder']
        shutil.copytree(code_folder, os.path.join(inner_folder, os.path.normpath(code_folder).split('/')[-1]))

        # 将配置文件重新写入 (替换脚本目录地址)
        config_content['folder'] = os.path.normpath(code_folder).split('/')[-1]
        with open(os.path.join(inner_folder, 'submit-config.yaml'), encoding='utf-8', mode='w') as fp:
          yaml.safe_dump(config_content, fp)
      return

    if has_config_file:
      # 进入到这里，说明已经指定ssh,local,k8s
      # 添加ssh,local,k8s的标准任务提交配置
      shutil.copy(args.config, os.path.join(os.environ['HOME'], '.config', 'antgo'))
    else:
      if args.ssh:
        # 生成ssh提交配置模板
        ssh_submit_config_file = os.path.join(os.path.dirname(__file__), 'script', 'ssh-submit-config.yaml')
        shutil.copy(ssh_submit_config_file, './')        
      else:
        # 生成自定义的提交配置模板
        submit_config_file = os.path.join(os.path.dirname(__file__), 'script', 'submit-config.yaml')
        shutil.copy(submit_config_file, './')    
    return

  ######################################### 后台监控服务 ################################################
  if action_name == 'server':
    os.system(f'nohup python3 {os.path.join(os.path.dirname(__file__), "ant", "client.py")} --port={args.port} --root={args.root} --ext-module={args.ext_module} > /tmp/antgo.server.log 2>&1 &')
    return
    # launch_server(args.port, args.root, args.ext_module)
    # return

  # 检查是否后台服务活跃
  if not get_client().alive():
    logging.warn('Antgo auto server not work. Please run "antgo server" to launch.')

  ######################################### 生成最小mvp ###############################################
  if action_name == 'create' and sub_action_name == 'mvp':
    # 在项目目录下生成mvp
    project_folder = os.path.join(os.path.dirname(__file__), 'resource', 'templates', 'mvp')
    generate_project_exp_example(project_folder, args.tgt, args.name)
    return

  ####################################################################################################
  # 执行指令
  if action_name in action_level_1:
    # 云端任务执行
    if args.cloud:
      # 检查项目环境
      if not check_project_environment(args):
        return

      # 检查项目和实验是否存在
      if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} hasnot create.')
        return

      project_info = {}      
      with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp)

      if args.exp not in project_info['exp']:
        logging.error(f'Exp {args.exp} not exist in project {args.project}.')
        return

      if not os.path.exists(args.exp):
        logging.error(f'Exp code not found in current folder.')
        return

      # 检查任务提交功能
      if args.ssh:
        ssh_submit_config_file = os.path.join(os.environ['HOME'], '.config', 'antgo', 'ssh-submit-config.yaml')
        if not os.path.exists(ssh_submit_config_file):
          logging.error('No ssh submit config.')
          return
      elif not args.local:
        submit_config_file = os.path.join(os.environ['HOME'], '.config', 'antgo', 'submit-config.yaml')    
        if not os.path.exists(submit_config_file):
            logging.error('No custom submit script config')
            return

      # 记录commit
      rep = git.Repo('./')     
      if not isinstance(project_info['exp'][args.exp], list):
        project_info['exp'][args.exp] = []
      
      exp_info = exp_basic_info()
      exp_info['exp'] = args.exp
      exp_info['id'] = time.strftime('%Y-%m-%dx%H-%M-%S',time.localtime(time.time()))
      exp_info['branch'] = rep.active_branch.name
      exp_info['commit'] = rep.active_branch.commit.name_rev
      exp_info['state'] = 'running'
      exp_info['stage'] = args.stage
      project_info['exp'][args.exp].append(
        exp_info
      )

      with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
        json.dump(project_info, fp)
      
      # 云端提交        
      sys_argv_cp.append(f'--id={exp_info["id"]}')
      filter_sys_argv_cp = []
      for t in sys_argv_cp:
        if t.startswith('--project'):
          continue
        if t.startswith('--root'):
          continue
        
        filter_sys_argv_cp.append(t)
      
      assert(args.root)
      sys_argv_cp = filter_sys_argv_cp
      sys_argv_cp.append(f'--root={args.root}/{args.project}')
      
      sys_argv_cmd = ' '.join(sys_argv_cp[1:])
      sys_argv_cmd = sys_argv_cmd.replace('--cloud', '')
      
      # step 1.1: 检查提交脚本配置
      if args.local:
        # 本地提交
        sys_argv_cmd = sys_argv_cmd.replace('--local', '')
        sys_argv_cmd = sys_argv_cmd.replace('  ', ' ')
        sys_argv_cmd = f'antgo {sys_argv_cmd}'
        local_submit_process_func(args.project, sys_argv_cmd, 0 if args.gpu_id == '' else len(args.gpu_id.split(',')), args.cpu, args.memory)  
      elif args.ssh:
        # ssh提交
        sys_argv_cmd = sys_argv_cmd.replace('--ssh', '')
        sys_argv_cmd = sys_argv_cmd.replace('  ', ' ')
        sys_argv_cmd = f'antgo {sys_argv_cmd}'        
        ssh_submit_process_func(args.project, sys_argv_cmd, 0 if args.gpu_id == '' else len(args.gpu_id.split(',')), args.cpu, args.memory)  
      else:
        # 自定义脚本提交
        sys_argv_cmd = sys_argv_cmd.replace('  ', ' ')
        sys_argv_cmd = f'antgo {sys_argv_cmd}'          
        custom_submit_process_func(args.project, sys_argv_cmd, 0 if args.gpu_id == '' else len(args.gpu_id.split(',')), args.cpu, args.memory)
      return
    # 本地任务执行
    elif args.project != '':
      # 可能，运行环境就是本地
      # 检查项目环境
      if not check_project_environment(args):
        return

      # 检查项目和实验是否存在
      if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
        logging.error(f'Project {args.project} hasnot create.')
        return

      project_info = {}      
      with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
        project_info = json.load(fp)

      if args.exp not in project_info['exp']:
        logging.error(f'Exp {args.exp} not exist in project {args.project}.')
        return
      
      if not os.path.exists(args.exp):
        logging.error(f'Exp code not found in current folder.')
        return

      # 记录commit
      rep = git.Repo('./')             
      if not isinstance(project_info['exp'][args.exp], list):
        project_info['exp'][args.exp] = []
      
      exp_info = exp_basic_info()
      exp_info['exp'] = args.exp
      exp_info['id'] = time.strftime('%Y-%m-%dx%H-%M-%S',time.localtime(time.time()))
      exp_info['branch'] = rep.active_branch.name
      exp_info['commit'] = rep.active_branch.commit.name_rev
      exp_info['state'] = 'running'
      exp_info['stage'] = args.stage
      project_info['exp'][args.exp].append(
        exp_info
      )

      with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
        json.dump(project_info, fp)

      # exp, exp:id
      args.id = exp_info['id']
    
    # 检查后端存储功能
    if args.root is not None and args.root != '':
      status = FileClient.infer_client(uri=args.root)
      if status is None:
        logging.error(f"Fail check backend storage {args.root}")
    
    # 执行任务
    auto_exp_name = f'{args.exp}.{args.id}' if args.id is not None else args.exp
    script_folder = os.path.join(os.path.dirname(__file__), 'script')
    if action_name == 'train':
      if args.checkpoint is None:
        args.checkpoint = ''
      if args.resume_from is None:
        args.resume_from = ''
        
      if args.gpu_id == '' or int(args.gpu_id.split(',')[0]) == -1:
        # cpu run
        # (1)安装;(2)数据准备;(3)运行
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint} --resume-from={args.resume_from}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={-1} --process=train --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'

        os.system(command_str)
      elif len(args.gpu_id.split(',')) == 1:
        # single gpu run
        # (1)安装;(2)数据准备;(3)运行
        gpu_id = args.gpu_id.split(',')[0]
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint} --resume-from={args.resume_from}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={gpu_id} --process=train --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'

        os.system(command_str)
      else:
        # multi gpu run
        # (1)安装;(2)数据准备;(3)运行
        gpu_num = len(args.gpu_id.split(','))
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint} --resume-from={args.resume_from}; bash launch.sh {args.exp}/main.py {gpu_num} --exp={auto_exp_name}  --process=train --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'
             
        os.system(command_str)
    elif action_name == 'activelearning':
      if args.checkpoint is None:
        args.checkpoint = ''

      if args.gpu_id == '' or int(args.gpu_id.split(',')[0]) == -1:
        # cpu run
        # (1)安装;(2)数据准备;(3)运行
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={-1} --process=activelearning --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'

        os.system(command_str)
      elif len(args.gpu_id.split(',')) == 1:
        # single gpu run
        # (1)安装;(2)数据准备;(3)运行
        gpu_id = args.gpu_id.split(',')[0]
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={gpu_id} --process=activelearning --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'

        os.system(command_str)
      else:
        # multi gpu run
        # (1)安装;(2)数据准备;(3)运行
        gpu_num = len(args.gpu_id.split(','))
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; bash launch.sh {args.exp}/main.py {gpu_num} --exp={auto_exp_name}  --process=activelearning --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.no_validate:
          command_str += ' --no-validate'
        if args.resume_from is not None:
          command_str += f' --resume-from={args.resume_from}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        if args.max_epochs > 0:
          command_str += f' --max-epochs={args.max_epochs}'

        os.system(command_str)
    elif action_name == 'eval':
      assert(args.checkpoint is not None)      
      # (1)安装;(2)数据准备;(3)运行
      if args.gpu_id == '' or int(args.gpu_id.split(',')[0]) == -1:
        # cpu run
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={-1} --process=test --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        os.system(command_str)
      elif len(args.gpu_id.split(',')) == 1:
        # single gpu run
        gpu_id = args.gpu_id.split(',')[0]
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; python3 {args.exp}/main.py --exp={auto_exp_name} --gpu-id={gpu_id} --process=test --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        os.system(command_str)
      else:
        # multi gpu run
        gpu_num = len(args.gpu_id.split(','))
        command_str = f'bash install.sh; python3 {script_folder}/data_prepare.py --exp={args.exp} --extra-config={args.extra_config} --config={args.config} --checkpoint={args.checkpoint}; bash launch.sh {args.exp}/main.py {gpu_num} --exp={auto_exp_name} --process=test --root={args.root} --extra-config={args.extra_config} --config={args.config}'
        if args.checkpoint is not None:
          command_str += f' --checkpoint={args.checkpoint}'
        os.system(command_str)
    elif action_name == 'export':
      assert(args.checkpoint is not None)
      os.system(f'bash install.sh; python3 {args.exp}/main.py --exp={auto_exp_name} --checkpoint={args.checkpoint} --process=export --root={args.root}')
  else:
    if action_name == 'create':
      if sub_action_name == 'project':
        if args.git is None or args.git == '':
          logging.error('Must set project git with --git')
          return

        # 自动提取git的后缀作为项目名称
        project_name = args.git.split('/')[-1].split('.')[0]

        if os.path.exists(os.path.join(config.AntConfig.task_factory,f'{project_name}.json')):
          logging.error(f'Project {project_name} has existed.')
          return

        # 加载项目默认模板
        with open(os.path.join(os.path.dirname(__file__), 'resource', 'templates', 'project.json'), 'r') as fp:
          project_config = json.load(fp)

        # 配置定制化项目
        project_config['name'] = project_name    # 项目名称
        project_config['type'] = args.type if args.type is not None else '' # 项目类型
        project_config['git'] = args.git      # 项目git地址
        project_config['image'] = args.image  # 项目需要的定制化的镜像
        project_config['auto'] = args.auto    # 项目是否进行自动化更新  （半监督\蒸馏\迁移）
        # 设置自动优化工具
        project_config['tool']['semi']['method'] = args.semi
        project_config['tool']['distillation']['method'] = args.distillation
        project_config['tool']['activelearning']['method'] = args.activelearning
        project_config['tool']['ensemble']['method'] = args.ensemble

        # 在本地存储项目信息
        with open(os.path.join(config.AntConfig.task_factory,f'{project_name}.json'), 'w') as fp:
          json.dump(project_config, fp)

        # 准备项目代码
        prepare_project_environment(args.git, args.branch, args.commit)
        pprint(project_config)
      elif sub_action_name == 'exp':
        # 检查项目环境
        if not check_project_environment(args):
          return

        # 获得基本项目信息
        if not os.path.exists(os.path.join(config.AntConfig.task_factory,f'{args.project}.json')):
          logging.error(f'Project {args.project} has not been created.')
          return
        project_info = {}
        with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'r') as fp:
          project_info = json.load(fp) 

        # 项目必须存在git
        assert(project_info['git'] != '')    
        
        # 创建项目实验
        assert(args.name is not None)
        if args.name in project_info['exp']:
          logging.error(f'Experiment {args.name} existed')
          return        

        if not os.path.exists(args.name):
          # 如果不存在实验目录，创建
          os.makedirs(args.name)

        if args.name not in project_info['exp']:
          project_info['exp'][args.name] = []

        with open(os.path.join(config.AntConfig.task_factory,f'{args.project}.json'), 'w') as fp:
          json.dump(project_info, fp)
    elif action_name == 'add':
      # expert, product, basline, train/label, train/unlabel, test, val
      project_add_action(sub_action_name, args)
    elif action_name == 'del':
      # expert, product, basline, train/label, train/unlabel, test, val
      project_del_action(sub_action_name, args)
    elif action_name == 'tool':
      # 工具相关
      if sub_action_name.startswith('extract'):
        # extract/videos, extract/images
        tool_func = getattr(tools, f'extract_from_{sub_action_name.split("/")[1]}', None)
        if tool_func is None:
          logging.error(f'Tool {sub_action_name} not exist.')
          return
        tool_func(
          args.src, 
          args.tgt, 
          frame_rate=args.frame_rate, 
          filter_prefix=args.prefix, 
          filter_suffix=args.suffix, 
          filter_ext=args.ext,
          feedback=args.feedback,
          num=args.num,
          shuffle=args.shuffle,
          max_size=args.max_size,
          ext_ratio=args.ext_ratio)
      elif sub_action_name.startswith('browser'):
        tool_func = getattr(tools, f'browser_{sub_action_name.split("/")[-1]}', None)
        if tool_func is None:
          logging.error(f'Tool {sub_action_name} not exist.')
          return
        tool_func(args.src, args.tags, args.white_users, args.feedback, args.user_input)
      elif sub_action_name.startswith('filter'):
        tool_func = getattr(tools, f'filter_by_{sub_action_name.split("/")[1]}', None)

        if tool_func is None:
          logging.error(f'Tool {sub_action_name} not exist.')
          return 
        
        tool_func(args.src, args.tgt, args.tags, args.no_tags)
        return
      elif sub_action_name.startswith('package'):
        tool_func = getattr(tools, f'package_to_{sub_action_name.split("/")[1]}', None)

        if tool_func is None:
          logging.error(f'Tool {sub_action_name} not exist.')
          return
        
        # src: json文件路径
        # tgt: 打包数据存放目录
        # prefix: 打包数据的文件前缀
        # num: 每个shard内样本条数
        tool_func(args.src, args.tgt, args.prefix, args.num)
      elif sub_action_name.startswith('download'):
        tool_func = getattr(tools, f'download_from_{sub_action_name.split("/")[1]}', None)

        if tool_func is None:
          logging.error(f'Tool {sub_action_name} not exist.')
          return
 
        tool_func(args.tgt, args.tags)
      elif sub_action_name.startswith('label'):
        if sub_action_name.split("/")[1] == 'start':
          tool_func = getattr(tools, f'label_{sub_action_name.split("/")[1]}', None)
          if tool_func is None:
            logging.error(f'Tool {sub_action_name} not exist.')
            return

          tool_func(args.src, args.tgt, args.tags, args.type, args.white_users, ignore_incomplete=args.ignore_incomplete, root=args.root, exp=args.exp)
          return
        else:
          tool_func = None
          if args.to:
            tool_func = getattr(tools, f'label_to_{sub_action_name.split("/")[1]}', None)
          else:
            tool_func = getattr(tools, f'label_from_{sub_action_name.split("/")[1]}', None)

          if tool_func is None:
            logging.error(f'Tool {sub_action_name} not exist.')
            return

          tool_func(args.src, args.tgt, prefix=args.prefix, tags=args.tags, ignore_incomplete=args.ignore_incomplete)
    elif action_name == 'show':
      show_action(sub_action_name, args)
    elif action_name == 'get':
      get_action(sub_action_name, args)
    elif action_name == 'update':
      update_project_config(sub_action_name, args)
    elif action_name == 'share':
      if sub_action_name == 'data':
        tools.share_data_func(args)


if __name__ == '__main__':
  main()



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 14:19
# @File    : io.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.pipeline.engine import *
import cv2
import os
import base64
import numpy as np
import requests
from antgo.utils.sample_gt import *
import imagesize


@register
class image_decode(object):
  def __init__(self, to_rgb=False) -> None:
    self.to_rgb = to_rgb
  
  def __call__(self, x):
    image = cv2.imread(x)
    if self.to_rgb:
      image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)    
    return image


@register
class image_decode_from_buffer(object):
  def __init__(self, to_rgb=False) -> None:
    self.to_rgb = to_rgb
  
  def __call__(self, x):
    data = np.frombuffer(x, dtype='uint8')
    image = cv2.imdecode(data, 1)  # BGR mode, but need RGB mode
    if self.to_rgb:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)        
    return image  
     

@register
def image_base64_decode(x):
  content = base64.b64decode(x)
  image = cv2.imdecode(np.frombuffer(content, np.uint8), cv2.IMREAD_COLOR)
  return image


@register
def image_download(image_url):
  pic = requests.get(image_url, timeout=20)
  image = cv2.imdecode(np.frombuffer(pic.content, np.uint8), cv2.IMREAD_COLOR)  
  return image
  
@register
def serialize_numpy(*args):
  serialize_result = []
  for v in args:
    if isinstance(v, np.ndarray):
      serialize_result.append(v.tolist())
    else:
      serialize_result.append(v)

  return tuple(serialize_result)

@register
class image_save(object):
  def __init__(self, folder, to_bgr=False):
    self.folder = folder
    self.to_bgr = to_bgr
    if not os.path.exists(self.folder):
      os.makedirs(self.folder)

    self.count = 0
    
  def __call__(self, x, file_path=None):
    if self.to_bgr:
      image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)    

    file_name = ''
    if file_path is not None:
      file_name = file_path.split('/')[-1].split('.')[0]
    else:
      file_name = f'frame_{self.count}'
      self.count += 1
    cv2.imwrite(os.path.join(self.folder, f'{file_name}.png'), x)


@register
class convert_gt_annos(object):
  def __init__(self, fields, replace_prefix, fill_lambda_fields=None, fill_constant_fields=None) -> None:
    self.fields = fields
    self.replace_prefix = replace_prefix
    self.fill_constant_fields = fill_constant_fields
    self.fill_lambda_fields =  fill_lambda_fields
    
  def __call__(self, *args):
    sgt = SampleGTTemplate()
    gt_info = sgt.get()
    for field, data in zip(self.fields, args):
      if field == 'image':
        gt_info['height'] = data.shape[0]
        gt_info['width'] = data.shape[1]
        continue
      
      if self.fill_lambda_fields is not None:
        if field in self.fill_lambda_fields:
          fill_field = self.fill_lambda_fields[field]['field']
          lambda_func = self.fill_lambda_fields[field]['func']
          gt_info[fill_field] = lambda_func(data)
          continue

      if field not in gt_info:
        continue
              
      if isinstance(gt_info[field], list):
        gt_info[field] = data.tolist()
      elif field in ['image_file', 'semantic_file']:
        if field == 'image_file':
          width, height = imagesize.get(data)
          gt_info['height'] = height
          gt_info['width'] = width
        
        gt_info[field] = data.replace(self.replace_prefix[0], self.replace_prefix[1])
        
      else:
        gt_info[field] = data
    
    if self.fill_constant_fields is not None:
      for k, v in self.fill_constant_fields.items():
        gt_info[k] = v
    return gt_info


from antgo.framework.helper.runner import get_dist_info
from antgo.framework.helper.fileio import *
import os


def running_flag(root):
    if root is None:
        return
    rank, _ = get_dist_info()        
    if rank == 0:
        file_client = \
                FileClient.infer_client(None, root)
        file_client.put(b'', f'{root}/RUNNING')


def finish_flag(root):
    if root is None:
        return        
    rank, _ = get_dist_info()
    
    if rank == 0:        
        file_client = \
                FileClient.infer_client(None, root)        
        file_client.put(b'', f'{root}/FINISH')


def stop_flag(root):
    if root is None:
        return        
    rank, _ = get_dist_info()
    
    if rank == 0:     
        file_client = \
                FileClient.infer_client(None, root)        
        file_client.put(b'', f'{root}/STOP')



# Copyright (c) OpenMMLab. All rights reserved.
from .anchor_generator import (AnchorGenerator, LegacyAnchorGenerator,
                               YOLOAnchorGenerator)
from .builder import (ANCHOR_GENERATORS, PRIOR_GENERATORS,
                      build_anchor_generator, build_prior_generator)
from .point_generator import MlvlPointGenerator, PointGenerator
from .utils import anchor_inside_flags, calc_region, images_to_levels

__all__ = [
    'AnchorGenerator', 'LegacyAnchorGenerator', 'anchor_inside_flags',
    'PointGenerator', 'images_to_levels', 'calc_region',
    'build_anchor_generator', 'ANCHOR_GENERATORS', 'YOLOAnchorGenerator',
    'build_prior_generator', 'PRIOR_GENERATORS', 'MlvlPointGenerator'
]
